import datatable as dt
import logging
import numpy as np
import pysam
import tqdm
import sys
import random
from bam_filter.utils import (
    create_empty_output_files,
    sort_keys_by_approx_weight,
    is_debug,
    get_arguments,
    check_tmp_dir_exists,
    handle_warning,
    create_output_files,
)
from multiprocessing import Pool, Manager, cpu_count
from functools import partial
import numpy.lib.recfunctions as rf
import gc
from collections import defaultdict
import os
import concurrent.futures
import math
import warnings
from bam_filter.sam_utils import check_bam_file

# import cProfile as prof
# import pstats

log = logging.getLogger("my_logger")


def initialize_subject_weights(data):
    if data.shape[0] > 0:
        # Add a new column for inverse sequence length
        data["s_W"] = 1 / data["slen"]

        # Calculate the sum of weights for each unique source
        sum_weights = np.zeros(int(np.max(data["source"])) + 1)
        np.add.at(sum_weights, data["source"], data["var"])

        # Calculate the normalized weights based on the sum for each query
        query_sum_var = sum_weights[data["source"]]
        data["prob"] = data["var"] / query_sum_var

        return data
    else:
        return None


def resolve_multimaps(data, scale=0.9, iters=10):
    current_iter = 0
    while True:
        progress_bar = tqdm.tqdm(
            total=9,
            desc=f"Iter {current_iter + 1}",
            unit=" step",
            disable=False,  # Replace with your logic or a boolean value
            leave=False,
            ncols=80,
        )
        log.debug(f"::: Iter: {current_iter + 1} - Getting scores")
        progress_bar.update(1)
        n_alns = data.shape[0]
        log.debug(f"::: Iter: {current_iter + 1} - Total alignment: {n_alns:,}")

        # Calculate the weights for each subject
        log.debug(f"::: Iter: {current_iter + 1} - Calculating weights...")
        progress_bar.update(1)
        subject_weights = np.zeros(int(np.max(data["subject"])) + 1)
        np.add.at(subject_weights, data["subject"], data["prob"])
        data["s_W"] = subject_weights[data["subject"]] / data["slen"]
        subject_weights = None

        log.debug(f"::: Iter: {current_iter + 1} - Calculating probabilities")
        progress_bar.update(1)
        # Calculate the alignment probabilities
        new_prob = data["prob"] * data["s_W"]
        log.debug("Calculating sum of probabilities")
        progress_bar.update(1)
        prob_sum = data["prob"] * data["s_W"]
        prob_sum_array = np.zeros(int(np.max(data["source"])) + 1)
        np.add.at(prob_sum_array, data["source"], prob_sum)
        prob_sum = None

        # data["prob_sum"] = prob_sum_array[data["source"]]
        data["prob"] = new_prob / prob_sum_array[data["source"]]
        prob_sum_array = None

        log.debug("Calculating query counts")
        progress_bar.update(1)
        # Calculate how many alignments are in each query
        # query_counts = np.zeros(int(np.max(data["source"])) + 1)
        # np.add.at(query_counts, data["source"], 1)

        query_counts = np.bincount(data["source"])

        log.debug("Calculating query counts array")
        progress_bar.update(1)
        # Use a separate array for query counts
        query_counts_array = np.zeros(int(np.max(data["source"])) + 1)
        np.add.at(
            query_counts_array,
            data["source"],
            query_counts[data["source"]],
        )

        log.debug(
            f"::: Iter: {current_iter + 1} - Calculating number of alignments per query"
        )
        progress_bar.update(1)
        data["n_aln"] = query_counts_array[data["source"]]

        log.debug("Calculating unique alignments")
        data["n_aln"] = query_counts_array[data["source"]]
        data_unique = data[data["n_aln"] == 1]
        n_unique = data_unique.shape[0]

        if n_unique == data.shape[0]:
            progress_bar.close()
            log.info("::: ::: No more multimapping reads. Early stopping.")
            return data
        data = data[(data["n_aln"] > 1) & (data["prob"] > 0)]

        # total_n_unique = np.sum(query_counts_array[data["source"]] <= 1)

        query_counts = None
        query_counts_array = None

        log.debug("Calculating max_prob")
        # Keep the ones that have a probability higher than the maximum scaled probability
        max_prob = np.zeros(int(np.max(data["source"])) + 1)
        np.maximum.at(max_prob, data["source"], data["prob"])

        data["max_prob"] = max_prob[data["source"]]
        data["max_prob"] = data["max_prob"] * scale
        # data["max_prob"] = max_prob[data["source"]]
        log.debug(
            f"::: Iter: {current_iter + 1} - Removing alignments with lower probability"
        )
        progress_bar.update(1)
        to_remove = np.sum(data["prob"] < data["max_prob"])

        data = data[data["prob"] >= data["max_prob"]]
        max_prob = None

        # Update the iteration count in the function call
        current_iter += 1
        data["iter"] = current_iter
        data_unique["iter"] = current_iter

        query_counts = np.bincount(data["source"])
        total_n_unique = np.sum(query_counts[data["source"]] <= 1)

        # data_unique["iter"] = current_iter

        # data = np.concatenate([data, data_unique])
        data = np.concatenate([data, data_unique])
        data_unique = None

        keep_processing = to_remove != 0
        log.debug(f"::: Iter: {current_iter} - Removed {to_remove:,} alignments")
        log.debug(f"::: Iter: {current_iter} - Total mapping queries: {n_unique:,}")
        log.debug(
            f"::: Iter: {current_iter} - New unique mapping queries: {total_n_unique:,}"
        )
        log.debug(f"::: Iter: {current_iter} - Alns left: {data.shape[0]:,}")
        progress_bar.update(1)
        progress_bar.close()
        log.info(
            f"::: Iter: {current_iter} - R: {to_remove:,} | U: {total_n_unique:,} | NU: {n_unique:,} | L: {data.shape[0]:,}"
        )
        log.debug(f"::: Iter: {current_iter} - done!")

        if iters > 0 and current_iter >= iters:
            log.info("::: ::: Reached maximum iterations. Stopping.")
            break
        elif not keep_processing:
            log.info("::: ::: No more alignments to remove. Stopping.")
            break
    return data


# def write_reassigned_bam(
#     bam, out_files, threads, entries, sort_memory="1G", min_read_ani=90
# ):
#     if out_files["bam_reassigned"] is not None:
#         out_bam = out_files["bam_reassigned"]
#     else:
#         out_bam = out_files["bam_reassigned_sorted"]

#     samfile = pysam.AlignmentFile(bam, "rb", threads=threads)
#     references = list(entries.keys())
#     refs_dict = {x: samfile.get_reference_length(x) for x in list(entries.keys())}

#     (ref_names, ref_lengths) = zip(*refs_dict.items())

#     refs_idx = {sys.intern(str(x)): i for i, x in enumerate(ref_names)}
#     if threads > 4:
#         write_threads = 4
#     else:
#         write_threads = threads

#     out_bam_file = pysam.AlignmentFile(
#         out_files["bam_reassigned_tmp"],
#         "wb",
#         referencenames=list(ref_names),
#         referencelengths=list(ref_lengths),
#         threads=write_threads,
#     )

#     for reference in tqdm.tqdm(
#         references,
#         total=len(references),
#         leave=False,
#         ncols=80,
#         desc="References processed",
#     ):
#         r_ids = entries[reference]
#         for aln in samfile.fetch(
#             reference=reference, multiple_iterators=False, until_eof=True
#         ):
#             # ani_read = (1 - ((aln.get_tag("NM") / aln.infer_query_length()))) * 100
#             if (aln.query_name, reference) in r_ids:
#                 aln.reference_id = refs_idx[aln.reference_name]
#                 out_bam_file.write(aln)
#     out_bam_file.close()


# def write_to_file(alns, out_bam_file):
#     for aln in tqdm.tqdm(alns, total=len(alns), leave=False, ncols=80, desc="Writing"):
#         out_bam_file.write(aln)


def write_to_file(alns, out_bam_file, header=None):
    for aln in alns:
        out_bam_file.write(pysam.AlignedSegment.fromstring(aln, header))


def process_references_batch(references, entries, bam, refs_idx, threads=4):
    alns = []
    if threads > 4:
        s_threads = 4
    else:
        s_threads = threads
    with pysam.AlignmentFile(bam, "rb", threads=s_threads) as samfile:
        for reference in references:
            r_ids = entries[reference]
            for aln in samfile.fetch(
                reference=reference, multiple_iterators=False, until_eof=True
            ):
                if (aln.query_name, reference) in r_ids:
                    aln.reference_id = refs_idx[aln.reference_name]
                    alns.append(aln.to_string())

    return alns


def write_reassigned_bam(
    bam,
    ref_counts,
    out_files,
    threads,
    entries,
    sort_memory="1G",
    sort_by_name=False,
    min_read_ani=90,
    min_read_length=30,
    disable_sort=False,
):
    # if out_files["bam_reassigned"] is not None:
    #     out_bam = out_files["bam_reassigned"]
    # else:
    #     out_bam = out_files["bam_reassigned_sorted"]
    out_bam = out_files["bam_reassigned"]
    if threads > 4:
        s_threads = 4
    else:
        s_threads = threads
    samfile = pysam.AlignmentFile(bam, "rb", threads=s_threads)
    references = list(entries.keys())
    refs_dict = {x: samfile.get_reference_length(x) for x in references}
    # get group reads
    header = samfile.header
    samfile.close()
    (ref_names, ref_lengths) = zip(*refs_dict.items())

    refs_idx = {sys.intern(str(x)): i for i, x in enumerate(ref_names)}
    if threads > 4:
        write_threads = 4
    else:
        write_threads = threads

    out_bam_file = pysam.AlignmentFile(
        out_files["bam_reassigned_tmp"],
        "wb",
        referencenames=list(ref_names),
        referencelengths=list(ref_lengths),
        threads=write_threads,
        header=header,
    )

    # num_cores should be multiple of the write_threads
    num_cores = threads // write_threads
    # num_cores = min(threads, cpu_count())
    # batch_size = len(references) // num_cores + 1  # Ensure non-zero batch size
    # batch_size = calc_chunksize(n_workers=num_cores, len_iterable=len(references))
    log.info("::: Creating reference chunks with uniform read amounts...")

    ref_chunks = sort_keys_by_approx_weight(
        input_dict=ref_counts, scale=1, num_cores=num_cores
    )
    num_cores = min(num_cores, len(ref_chunks))
    log.info(f"::: Using {num_cores} processes to write {len(ref_chunks)} chunk(s)")

    with Manager() as manager:
        # Use Manager to create a read-only proxy for the dictionary
        entries = manager.dict(dict(entries))

        with concurrent.futures.ProcessPoolExecutor(max_workers=num_cores) as executor:
            # Use ProcessPoolExecutor to parallelize the processing of references in batches
            futures = []
            for batch_references in tqdm.tqdm(
                ref_chunks,
                total=len(ref_chunks),
                desc="Submitted batches",
                unit="batch",
                leave=False,
                ncols=80,
                disable=is_debug(),
            ):
                future = executor.submit(
                    process_references_batch, batch_references, entries, bam, refs_idx
                )
                futures.append(future)  # Store the future

            # Use a while loop to continuously check for completed futures
            log.info("::: Collecting batches...")

            completion_progress_bar = tqdm.tqdm(
                total=len(futures),
                desc="Completed",
                unit="batch",
                leave=False,
                ncols=80,
                disable=is_debug(),
            )
            completed_count = 0

            # Use as_completed to iterate over completed futures as they become available
            for completed_future in concurrent.futures.as_completed(futures):
                alns = completed_future.result()
                write_to_file(alns=alns, out_bam_file=out_bam_file, header=header)

                # Update the progress bar for each completed write
                completion_progress_bar.update(1)
                completed_count += 1
                completed_future.cancel()  # Cancel the future to free memory
                gc.collect()  # Force garbage collection

            completion_progress_bar.close()
    out_bam_file.close()
    entries = None
    gc.collect()
    # prof.disable()
    # # print profiling output
    # stats = pstats.Stats(prof).strip_dirs().sort_stats("tottime")
    # stats.print_stats(5)  # top 10 rows
    if not disable_sort:
        log.info("::: ::: Sorting BAM file...")
        if threads > 4:
            s_threads = 4
        else:
            s_threads = threads
        if sort_by_name:
            log.info("::: ::: Sorting by name...")
            pysam.sort(
                "-n",
                "-@",
                str(s_threads),
                "-m",
                str(sort_memory),
                "-o",
                out_bam,
                out_files["bam_reassigned_tmp"],
            )
        else:
            pysam.sort(
                "-@",
                str(s_threads),
                "-m",
                str(sort_memory),
                "-o",
                out_bam,
                out_files["bam_reassigned_tmp"],
            )

        logging.info("BAM index not found. Indexing...")
        save = pysam.set_verbosity(0)
        if threads > 4:
            s_threads = 4
        else:
            s_threads = threads
        samfile = pysam.AlignmentFile(out_bam, "rb", threads=s_threads)
        chr_lengths = []
        for chrom in samfile.references:
            chr_lengths.append(samfile.get_reference_length(chrom))
        max_chr_length = np.max(chr_lengths)
        pysam.set_verbosity(save)
        samfile.close()

        if max_chr_length > 536870912:
            logging.info("A reference is longer than 2^29, indexing with csi")
            pysam.index(
                "-c",
                "-@",
                str(threads),
                out_bam,
            )
        else:
            pysam.index(
                "-@",
                str(threads),
                out_bam,
            )

        os.remove(out_files["bam_reassigned_tmp"])
    else:
        logging.info("Skipping BAM file sorting...")
        os.rename(out_files["bam_reassigned_tmp"], out_bam)


# def calculate_alignment_score(identity, read_length):
#     return (identity / math.log(read_length)) * math.sqrt(read_length)

# Values from:
# https://www.ncbi.nlm.nih.gov/IEB/ToolBox/CPP_DOC/lxr/source/src/algo/blast/core/blast_stat.c
# def calculate_alignment_score(
#         num_matches,
#         num_mismatches,
#         num_gaps,
#         gap_extensions,
#         match_reward,
#         mismatch_penalty,
#         gap_open_penalty,
#         gap_extension_penalty,
#         lambda_value,
#         K_value
#     ):
#     # Calculate the raw alignment score
#     S = (num_matches * match_reward) - (num_mismatches * mismatch_penalty) - (num_gaps * gap_open_penalty) - (gap_extensions * gap_extension_penalty)

#     # Calculate the approximate bit score
#     bit_score = (lambda_value * S - math.log(K_value)) / math.log(2)

#     return bit_score

# def get_bam_data(
#         parms,
#         ref_lengths=None,
#         percid=90,
#         min_read_length=30,
#         threads=1,
#         match_reward=1,
#         mismatch_penalty=-1,
#         gap_open_penalty=1,
#         gap_extension_penalty=2,
#         lambda_value=1.02,
#         K_value=0.21
#     ):
#     bam, references = parms
#     dt.options.progress.enabled = False
#     dt.options.progress.clear_on_success = True
#     if threads > 1:
#         dt.options.nthreads = threads - 1
#     else:
#         dt.options.nthreads = 1
#     if threads > 4:
#         s_threads = 4
#     else:
#         s_threads = threads
#     samfile = pysam.AlignmentFile(bam, "rb", threads=s_threads)

#     results = []
#     reads = set()
#     refs = set()
#     empty_df = 0

#     for reference in references:
#         if ref_lengths is None:
#             reference_length = int(samfile.get_reference_length(reference))
#         else:
#             reference_length = int(ref_lengths[reference])
#         aln_data = []
#         for aln in samfile.fetch(
#             contig=reference, multiple_iterators=False, until_eof=True
#         ):
#             # AS_value = aln.get_tag("AS") if aln.has_tag("AS") else None
#             # XS_value = aln.get_tag("XS") if aln.has_tag("XS") else None
#             query_length = (
#                 aln.query_length if aln.query_length != 0 else aln.infer_query_length()
#             )

#             if query_length < min_read_length:
#                 continue

#             pident = (1 - ((aln.get_tag("NM") / query_length))) * 100
#             num_mismatches = aln.get_tag("NM")
#             num_matches = query_length - num_mismatches
#             num_gaps = aln.get_tag("XO") if aln.has_tag("XO") else 0
#             gap_extensions = aln.get_tag("XG") if aln.has_tag("XG") else 0
#             if pident >= percid:
#                 var = calculate_alignment_score(num_matches=num_matches, num_mismatches=num_mismatches, num_gaps=num_gaps, gap_extensions=gap_extensions, match_reward=match_reward, mismatch_penalty=mismatch_penalty, gap_open_penalty=gap_open_penalty, gap_extension_penalty=gap_extension_penalty, lambda_value=lambda_value, K_value=K_value)
#                 aln_data.append(
#                     (
#                         aln.query_name,
#                         aln.reference_name,
#                         var,
#                         reference_length,
#                     )
#                 )
#                 reads.add(aln.query_name)
#                 refs.add(aln.reference_name)
#         # remove duplicates
#         # check if aln_data is not empty
#         if len(aln_data) > 0:
#             aln_data_dt = dt.Frame(aln_data)
#             # "queryId", "subjectId", "bitScore", "slen"
#             aln_data_dt.names = ["queryId", "subjectId", "var", "slen"]
#             # remove duplicates and keep the ones with the largest score
#             aln_data_dt = aln_data_dt[
#                 :1, :, dt.by(dt.f.queryId, dt.f.subjectId), dt.sort(-dt.f.var)
#             ]
#             results.append(aln_data_dt)
#         else:
#             results.append(dt.Frame())
#             empty_df += 1
#     samfile.close()
#     return (dt.rbind(results), reads, refs, empty_df)


def calculate_alignment_score(
    num_matches,
    num_mismatches,
    num_gaps,
    gap_extensions,
    match_reward,
    mismatch_penalty,
    gap_open_penalty,
    gap_extension_penalty,
    precomputed_factor,  # This is lambda_value * match_reward / math.log(2)
    precomputed_log_K,  # This is math.log(K_value) / math.log(2)
):
    # Calculate the raw alignment score with reduced arithmetic operations
    S = (
        (num_matches * match_reward)
        - (num_mismatches * mismatch_penalty)
        - (num_gaps * gap_open_penalty)
        - (gap_extensions * gap_extension_penalty)
    )

    # Use precomputed factors to calculate the approximate bit score
    bit_score = precomputed_factor * S - precomputed_log_K

    return bit_score


def get_bam_data(
    parms,
    ref_lengths=None,
    percid=90,
    min_read_length=30,
    threads=1,
    match_reward=1,
    mismatch_penalty=-1,
    gap_open_penalty=1,
    gap_extension_penalty=2,
    lambda_value=1.02,
    K_value=0.21,
):
    # Precompute factors for the score calculation to avoid redundant computation
    precomputed_factor = lambda_value * match_reward / math.log(2)
    precomputed_log_K = math.log(K_value) / math.log(2)

    bam, references = parms
    dt.options.progress.enabled = False
    dt.options.progress.clear_on_success = True
    dt.options.nthreads = max(1, threads - 1)
    s_threads = min(4, threads)
    samfile = pysam.AlignmentFile(bam, "rb", threads=s_threads)

    results = []
    reads = set()
    refs = set()
    empty_df = 0

    for reference in references:
        reference_length = (
            int(samfile.get_reference_length(reference))
            if ref_lengths is None
            else int(ref_lengths[reference])
        )
        aln_data = []
        for aln in samfile.fetch(
            contig=reference, multiple_iterators=False, until_eof=True
        ):
            query_length = (
                aln.query_length if aln.query_length != 0 else aln.infer_query_length()
            )

            if query_length >= min_read_length:
                num_mismatches = aln.get_tag("NM")
                pident = (1 - (num_mismatches / query_length)) * 100
                if pident >= percid:
                    num_matches = query_length - num_mismatches
                    num_gaps = aln.get_tag("XO") if aln.has_tag("XO") else 0
                    gap_extensions = aln.get_tag("XG") if aln.has_tag("XG") else 0

                    bit_score = calculate_alignment_score(
                        num_matches,
                        num_mismatches,
                        num_gaps,
                        gap_extensions,
                        match_reward,
                        mismatch_penalty,
                        gap_open_penalty,
                        gap_extension_penalty,
                        precomputed_factor,
                        precomputed_log_K,
                    )
                    aln_data.append(
                        (
                            aln.query_name,
                            aln.reference_name,
                            bit_score,
                            reference_length,
                        )
                    )
                    reads.add(aln.query_name)
                    refs.add(aln.reference_name)

        if aln_data:
            aln_data_dt = dt.Frame(
                aln_data, names=["queryId", "subjectId", "bitScore", "slen"]
            )
            aln_data_dt = aln_data_dt[
                :1, :, dt.by(dt.f.queryId, dt.f.subjectId), dt.sort(-dt.f.bitScore)
            ]
            results.append(aln_data_dt)
        else:
            empty_df += 1

    samfile.close()
    combined_results = dt.rbind(results)
    return (combined_results, reads, refs, empty_df)


def reassign_reads(
    bam,
    out_files,
    match_reward,
    mismatch_penalty,
    gap_open_penalty,
    gap_extension_penalty,
    lambda_value,
    K_value,
    reference_lengths=None,
    threads=1,
    min_read_count=1,
    min_read_ani=90,
    min_read_length=30,
    reassign_iters=25,
    reassign_scale=0.9,
    sort_memory="4G",
    sort_by_name=False,
    disable_sort=False,
):
    dt.options.progress.enabled = True
    dt.options.progress.clear_on_success = True
    if threads > 1:
        dt.options.nthreads = threads - 1
    else:
        dt.options.nthreads = 1

    log.info("::: Loading BAM file")
    save = pysam.set_verbosity(0)
    if threads > 4:
        s_threads = 4
    else:
        s_threads = threads
    samfile = pysam.AlignmentFile(bam, "rb", threads=s_threads)

    references = samfile.references

    pysam.set_verbosity(save)

    if reference_lengths is not None:
        ref_len_dt = dt.fread(reference_lengths)
        ref_len_dt.names = ["subjectId", "slen"]
        # convert to dict
        ref_len_dict = dict(
            zip(ref_len_dt["subjectId"].to_list()[0], ref_len_dt["slen"].to_list()[0])
        )
        # check if the dataframe contains all the References in the BAM file
        if not set(references).issubset(set(ref_len_dict.keys())):
            logging.error(
                "The BAM file contains references not found in the reference lengths file"
            )
            sys.exit(1)
    else:
        ref_len_dict = None

    total_refs = samfile.nreferences
    log.info(f"::: Found {total_refs:,} reference sequences")
    # logging.info(f"Found {samfile.mapped:,} alignments")
    log.info(f"::: Removing references with less than {min_read_count} reads...")
    references_m = {
        chrom.contig: chrom.mapped
        for chrom in samfile.get_index_statistics()
        if chrom.mapped >= min_read_count
    }

    references = list(references_m.keys())

    if len(references) == 0:
        log.warning("::: No reference sequences with alignments found in the BAM file")
        create_empty_output_files(out_files)
        sys.exit(0)

    log.info(f"::: Keeping {len(references):,} references")

    log.info("::: Creating reference chunks with uniform read amounts...")
    # ify the number of chunks
    ref_chunks = sort_keys_by_approx_weight(
        input_dict=references_m, scale=1, num_cores=threads, verbose=False
    )
    log.info(f"::: ::: Created {len(ref_chunks):,} chunks")
    ref_chunks = random.sample(ref_chunks, len(ref_chunks))

    dt.options.progress.enabled = False
    dt.options.progress.clear_on_success = True
    dt.options.nthreads = 1

    parms = list(zip([bam] * len(ref_chunks), ref_chunks))

    log.info("::: Extracting reads from BAM file...")
    if is_debug():
        data = list(
            tqdm.tqdm(
                map(
                    partial(
                        get_bam_data,
                        ref_lengths=ref_len_dict,
                        percid=min_read_ani,
                        min_read_length=min_read_length,
                        match_reward=match_reward,
                        mismatch_penalty=mismatch_penalty,
                        gap_open_penalty=gap_open_penalty,
                        gap_extension_penalty=gap_extension_penalty,
                        lambda_value=lambda_value,
                        K_value=K_value,
                        threads=4,
                    ),
                    parms,
                    chunksize=1,
                ),
                total=len(parms),
                leave=False,
                ncols=80,
                desc="Chunks processed",
            )
        )
    else:
        p = Pool(
            threads,
        )
        data = list(
            tqdm.tqdm(
                p.imap_unordered(
                    partial(
                        get_bam_data,
                        ref_lengths=ref_len_dict,
                        percid=min_read_ani,
                        min_read_length=min_read_length,
                        match_reward=match_reward,
                        mismatch_penalty=mismatch_penalty,
                        gap_open_penalty=gap_open_penalty,
                        gap_extension_penalty=gap_extension_penalty,
                        lambda_value=lambda_value,
                        K_value=K_value,
                        threads=4,
                    ),
                    parms,
                    chunksize=1,
                ),
                total=len(parms),
                leave=False,
                ncols=80,
                desc="Chunks processed",
            )
        )

        p.close()
        p.join()

    dt.options.progress.enabled = True
    dt.options.progress.clear_on_success = True
    if threads > 1:
        dt.options.nthreads = threads - 1
    else:
        dt.options.nthreads = 1

    log.info("::: Collecting results...")
    reads = list()
    refs = list()
    empty_df = 0
    # new_data = list()
    # for i in tqdm.tqdm(range(len(data)), total=len(data), leave=False, ncols=80):
    #     empty_df += data[i][3]
    #     reads.extend(list(data[i][1]))
    #     refs.extend(list(data[i][2]))
    #     data[i] = data[i][0]
    new_data = list()
    for i in tqdm.tqdm(range(len(data)), total=len(data), leave=False, ncols=80):
        empty_df += data[i][3]
        reads.extend(list(data[i][1]))
        refs.extend(list(data[i][2]))

        # Check if the frame has more than 2 billion rows
        if data[i][0].nrows > 2e9:
            # Calculate the number of chunks needed
            num_chunks = (data[i][0].nrows // 1e9) + (data[i][0].nrows % 1e9 > 0)
            log.warning(
                f"Frame has more than 2 billion rows. Splitting into {num_chunks:,} chunks..."
            )

            chunks = []

            # Create chunks of 1 billion rows each
            for chunk_idx in range(num_chunks):
                start = int(chunk_idx * 1e9)
                end = int(min((chunk_idx + 1) * 1e9, data[i][0].nrows))
                chunks.append(data[i][0][start:end, :])

            # Substitute data[i] with the first chunk
            data[i] = chunks[0]

            # Append the rest of the chunks to new_data
            for chunk in chunks[1:]:
                new_data.append(chunk)
        else:
            # If the frame is not larger than 2 billion rows, keep it as is
            data[i] = data[i][0]

    # Combine data with new_data
    data.extend(new_data)
    log.info(f"::: ::: Removed {empty_df:,} references without alignments")

    # data = dt.rbind([x for x in data])

    # log.info("::: Indexing references...")
    # refs = dt.Frame(list(set(refs)))
    # refs.names = ["subjectId"]
    # refs["sidx"] = dt.Frame(list(range(refs.shape[0])))
    # refs.key = "subjectId"

    # log.info("::: Indexing reads...")
    # reads = dt.Frame(list(set(reads)))
    # reads.names = ["queryId"]
    # reads["qidx"] = dt.Frame([(i + refs.shape[0]) for i in range(reads.shape[0])])
    # reads.key = "queryId"

    # log.info("::: Combining data...")
    # data = data[:, :, dt.join(reads)]
    # data = data[:, :, dt.join(refs)]

    # Initialize refs DataFrame
    log.info("::: Indexing references...")
    refs = dt.Frame(list(set(refs)))
    refs.names = ["subjectId"]
    refs["sidx"] = dt.Frame(list(range(refs.shape[0])))
    refs.key = "subjectId"

    # Initialize reads DataFrame
    log.info("::: Indexing reads...")
    reads = dt.Frame(list(set(reads)))
    reads.names = ["queryId"]
    reads["qidx"] = dt.Frame([(i + refs.shape[0]) for i in range(reads.shape[0])])
    reads.key = "queryId"
    n_alns_0 = 0
    # Loop through each DataFrame in the list and update it with the joined version
    log.info("::: Combining data...")
    # for i, x in tqdm.tqdm(
    #     enumerate(data),
    #     total=len(data),
    #     desc="Processing batches",
    #     unit="batch",
    #     disable=is_debug(),
    #     leave=False,
    #     ncols=80,
    # ):
    #     # Perform join with reads and then refs
    #     x = x[:, :, dt.join(reads)]
    #     x = x[:, :, dt.join(refs)]
    #     n_alns_0 += x.shape[0]
    #     del x["queryId"]
    #     del x["subjectId"]
    #     x = x[:, [dt.f.qidx, dt.f.sidx, dt.f.var, dt.f.slen]].to_numpy()
    #     # Substitute the original DataFrame with the joined version in the list
    #     data[i] = x

    # Calculate the total number of rows in advance
    total_rows = sum(
        x.shape[0] for x in data
    )  # This assumes `data` is a list of DataFrames/NumPy arrays

    # Assuming all `x` arrays have the same number of columns after processing, use the first one to determine this
    # IMPORTANT: This line needs to be executed before the loop and assumes all `x` arrays are similar after processing
    num_columns = 4  # Adjust based on your actual data structure

    # Preallocate the NumPy array
    mat = np.empty(
        (total_rows, num_columns), dtype=np.float32
    )  # Adjust dtype as necessary

    current_index = 0
    for i, x in tqdm.tqdm(
        enumerate(data),
        total=len(data),
        desc="Processing batches",
        unit="batch",
        disable=is_debug(),
        leave=False,
        ncols=80,
    ):
        # Perform join with reads and then refs
        if x.shape[0] > 0:
            x = x[:, :, dt.join(reads)]
            x = x[:, :, dt.join(refs)]
            n_alns_0 += x.shape[0]

            # Process `x` as before, but directly update `mat`
            x_processed = x[
                :, [dt.f.qidx, dt.f.sidx, dt.f.bitScore, dt.f.slen]
            ].to_numpy()
            num_rows = x_processed.shape[0]

            # Fill the preallocated array
            mat[current_index : current_index + num_rows, :] = x_processed

            # Update the current index
            current_index += num_rows

    # After the loop, `mat` is already the concatenated array, so there's no need for further concatenation or conversion.
    data = None  # Free the memory if `data` is no longer needed

    # Log the final stats
    n_reads_0 = reads.shape[0]
    n_refs_0 = refs.shape[0]
    log.info(
        f"::: References: {n_refs_0:,} | Reads: {n_reads_0:,} | Alignments: {n_alns_0:,}"
    )

    # After the loop, use dt.rbind() to combine all the DataFrames in the list
    # data = dt.rbind([x for x in data])

    # del data["queryId"]
    # del data["subjectId"]
    # n_alns_0 = data.shape[0]
    # n_reads_0 = reads.shape[0]
    # n_refs_0 = refs.shape[0]
    # log.info(
    #     f"::: References: {n_refs_0:,} | Reads: {n_reads_0:,} | Alignments: {n_alns_0:,}"
    # )

    # log.info("::: Allocating data...")
    # # # mat = data[:, [dt.f.qidx, dt.f.sidx, dt.f.var, dt.f.slen]].to_numpy()
    # # mat = np.vstack(data)
    # # data = None

    # # Create a zeros array with the same number of rows as 'mat'
    # zeros_array = np.zeros((mat.shape[0], 5))

    # # Stack the zeros_array with the original 'mat'
    # m = np.column_stack([mat, zeros_array])
    # zeros_array = None

    # dtype = np.dtype(
    #     [
    #         ("source", "int"),
    #         ("subject", "int"),
    #         ("var", "float"),
    #         ("slen", "int"),
    #         ("s_W", "float"),
    #         ("prob", "float"),
    #         ("iter", "int"),
    #         ("n_aln", "int"),
    #         ("max_prob", "float"),
    #     ]
    # )

    # # Convert the unstructured array to structured array
    # m = rf.unstructured_to_structured(m, dtype)
    # gc.collect()

    log.info("::: Allocating data...")

    # Define the dtype for the structured array
    dtype = np.dtype(
        [
            ("source", "int"),
            ("subject", "int"),
            ("var", "float"),
            ("slen", "int"),
            (
                "s_W",
                "float",
            ),  # This and following fields are initialized to 0 or a default value
            ("prob", "float"),
            ("iter", "int"),
            ("n_aln", "int"),
            ("max_prob", "float"),
        ]
    )

    # Initialize the structured array with zeros directly
    m = np.zeros(mat.shape[0], dtype=dtype)
    m["source"] = mat[:, 0]
    m["subject"] = mat[:, 1]
    m["var"] = mat[:, 2]
    m["slen"] = mat[:, 3]

    # Force a garbage collection to free up memory from any intermediate arrays that are no longer needed
    gc.collect()

    log.info("::: Initializing data structures...")
    init_data = initialize_subject_weights(m)
    if reassign_iters > 0:
        log.info(f"::: Reassigning reads with {reassign_iters} iterations")
    else:
        log.info("::: Reassigning reads until convergence")
    no_multimaps = resolve_multimaps(
        init_data, iters=reassign_iters, scale=reassign_scale
    )

    n_reads = len(list(set(no_multimaps["source"])))
    n_refs = len(list(set(no_multimaps["subject"])))
    n_alns = no_multimaps.shape[0]
    log.info(
        f"::: References: {n_refs:,} | Reads: {n_reads:,} | Alignments: {n_alns:,}"
    )
    log.info(
        f'::: Unique mapping reads: {no_multimaps[no_multimaps["n_aln"] == 1].shape[0]:,} | Multimapping reads: {no_multimaps[no_multimaps["n_aln"] > 1].shape[0]:,}'
    )

    # add this to the array
    # no_multimaps["n_aln"] = subject_counts_array[no_multimaps["subject"]]

    # log.info(f"::: Removing references with less than {min_read_count} reads...")
    # no_multimaps = no_multimaps[no_multimaps["n_aln"] >= min_read_count]
    # log.info(f"{no_multimaps.shape[0]:,} alignments left")
    log.info("::: Mapping back indices...")
    if threads > 1:
        dt.options.nthreads = threads - 1
    else:
        dt.options.nthreads = 1

    g = dt.Frame(no_multimaps["source"])
    g.names = ["qidx"]
    reads.key = "qidx"
    q = g[:, :, dt.join(reads)]

    g = dt.Frame(no_multimaps["subject"])
    g.names = ["sidx"]
    refs.key = "sidx"
    s = g[:, :, dt.join(refs)]

    log.info("::: Calculating reads per subject...")
    # count how many alignments are in each subjectId
    s_c = s[:, dt.count(dt.f.subjectId), dt.by(dt.f.subjectId)]
    s_c.names = ["subjectId", "counts"]
    references_m = dict()
    log.info(f"::: Removing references with less than {min_read_count:,}...")
    for i, k in zip(s_c[:, "subjectId"].to_list()[0], s_c[:, "counts"].to_list()[0]):
        if k >= min_read_count:
            references_m[i] = k
    log.info(f"::: ::: Keeping {len(references_m):,} references")
    s_c = None
    if len(references_m) == 0:
        log.warning("::: No reference sequences with alignments found in the BAM file")
        create_empty_output_files(out_files)
        sys.exit(0)
    # convert columns queryId from q and subjectId from s to a tuple
    log.info("::: Creating filtered set...")
    entries = defaultdict(set)
    q_query_ids = q[:, "queryId"].to_list()[0]
    s_subject_ids = s[:, "subjectId"].to_list()[0]

    for query_id, subject_id in zip(q_query_ids, s_subject_ids):
        if subject_id in references_m:
            entries[subject_id].add((query_id, subject_id))
    no_multimaps = None
    q = None
    s = None
    q_query_ids = None
    s_subject_ids = None
    gc.collect()
    log.info("::: Writing to BAM file...")
    write_reassigned_bam(
        bam=bam,
        ref_counts=references_m,
        out_files=out_files,
        threads=threads,
        entries=entries,
        sort_memory=sort_memory,
        min_read_ani=min_read_ani,
        disable_sort=disable_sort,
    )


def reassign(args):
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(levelname)s ::: %(asctime)s ::: %(message)s",
        datefmt="%H:%M:%S",
    )

    args = get_arguments()
    bam = args.bam
    tmp_dir = check_tmp_dir_exists(args.tmp_dir)
    log.info("Temporary directory: %s", tmp_dir.name)
    out_files = create_output_files(
        prefix=args.prefix,
        bam=args.bam,
        tmp_dir=tmp_dir,
        mode="reassign",
        bam_reassigned=args.bam_reassigned,
    )
    sorted_bam = bam.replace(".bam", ".bf-sorted.bam")
    bam = check_bam_file(
        bam=args.bam,
        threads=args.threads,
        reference_lengths=args.reference_lengths,
        sort_memory=args.sort_memory,
    )
    if bam is None:
        logging.warning("No reference sequences with alignments found in the BAM file")
        create_empty_output_files(out_files)
        sys.exit(0)

    logging.getLogger("my_logger").setLevel(
        logging.DEBUG if args.debug else logging.INFO
    )

    if args.debug:
        warnings.showwarning = handle_warning
    else:
        warnings.filterwarnings("ignore")
    logging.info("Resolving multi-mapping reads...")
    reassign_reads(
        bam=bam,
        threads=args.threads,
        reference_lengths=args.reference_lengths,
        min_read_count=args.min_read_count,
        min_read_ani=args.min_read_ani,
        min_read_length=args.min_read_length,
        reassign_iters=args.reassign_iters,
        reassign_scale=args.reassign_scale,
        sort_memory=args.sort_memory,
        sort_by_name=args.sort_by_name,
        out_files=out_files,
        match_reward=args.match_reward,
        mismatch_penalty=args.mismatch_penalty,
        gap_open_penalty=args.gap_open_penalty,
        gap_extension_penalty=args.gap_extension_penalty,
        lambda_value=args.lambda_value,
        K_value=args.K_value,
        disable_sort=args.disable_sort,
    )
    # check if sorted BAM file exists, if yes remove it
    if os.path.exists(sorted_bam):
        os.remove(sorted_bam)
    # check if sorted BAM index file exists, if yes remove it
    if os.path.exists(sorted_bam + ".bai"):
        os.remove(sorted_bam + ".bai")
    elif os.path.exists(sorted_bam + ".csi"):
        os.remove(sorted_bam + ".csi")
    log.info("Done!")



import argparse
import sys
import gzip
import os
import shutil
import logging
import pandas as pd
from multiprocessing import Pool
from functools import partial
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from os import devnull
import tqdm
from bam_filter import __version__
import time
from itertools import chain
import numpy as np
from pathlib import Path
import pysam
import tempfile


log = logging.getLogger("my_logger")
log.setLevel(logging.INFO)
timestr = time.strftime("%Y%m%d-%H%M%S")


def handle_warning(message, category, filename, lineno, file=None, line=None):
    print("A warning occurred:")
    print(message)
    print("Do you wish to continue?")

    while True:
        response = input("y/n: ").lower()
        if response not in {"y", "n"}:
            print("Not understood.")
        else:
            break

    if response == "n":
        raise category(message)


# Check if the temporary directory exists, if not, create it
def check_tmp_dir_exists(tmpdir):
    if tmpdir is None:
        tmpdir = tempfile.TemporaryDirectory(dir=os.getcwd())
    else:
        if not os.path.exists(tmpdir):
            log.error(f"Temporary directory {tmpdir} does not exist")
            exit(1)
        tmpdir = tempfile.TemporaryDirectory(dir=os.path.abspath(tmpdir))
        # Check if tmpdir has more than 107 characters
        if len(tmpdir.name) > 107:
            log.error(f"Temporary directory {tmpdir.name} has more than 107 characters")
            exit(1)
    return tmpdir


def is_debug():
    return logging.getLogger("my_logger").getEffectiveLevel() == logging.DEBUG


def refine_chunks(chunks, input_dict, target_weight):
    # Calculate the current weight of each chunk
    chunk_weights = [sum(input_dict[key] for key in chunk) for chunk in chunks]

    while max(chunk_weights) - min(chunk_weights) > target_weight:
        # Find the chunk with the maximum weight
        src_chunk = chunk_weights.index(max(chunk_weights))

        # Find the chunk with the minimum weight
        dest_chunk = chunk_weights.index(min(chunk_weights))

        # Find the key in the source chunk with the maximum weight
        max_weight_key = max(chunks[src_chunk], key=lambda key: input_dict[key])

        # Move the key with maximum weight from the source to the destination chunk
        chunks[src_chunk].remove(max_weight_key)
        chunks[dest_chunk].append(max_weight_key)

        # Recalculate chunk weights
        chunk_weights = [sum(input_dict[key] for key in chunk) for chunk in chunks]

    # Remove empty chunks
    chunks = [chunk for chunk in chunks if chunk]

    return chunks


def sort_keys_by_approx_weight(
    input_dict, scale=1, num_cores=1, refinement_steps=10, verbose=False
):
    if scale == 0:
        raise ValueError("Scale cannot be zero.")

    # Calculate the target weight for each chunk
    target_weight = scale * int(max(input_dict.values()))

    # Determine the initial number of chunks based on the number of cores
    # num_chunks = num_cores * scale
    num_chunks = (((sum(input_dict.values()) // target_weight)) // num_cores) + 1
    if num_chunks < num_cores:
        num_chunks = num_cores
    # Sort keys by their weights in descending order
    sorted_keys = sorted(input_dict, key=lambda k: input_dict[k], reverse=True)

    # Initialize chunks
    chunks = [[] for _ in range(num_chunks)]
    total_weight = [0] * num_chunks

    # Create a progress bar
    progress_bar = tqdm.tqdm(
        total=len(sorted_keys),
        desc="Distributing keys",
        unit="k",
        unit_scale=True,
        unit_divisor=1000,
        disable=False,
        leave=False,
        ncols=80,
    )

    # Distribute keys into chunks with weights close to target_weight
    for key in sorted_keys:
        # Find the chunk with the least total weight
        min_chunk_index = min(range(num_chunks), key=lambda i: total_weight[i])

        # If adding the key doesn't exceed the target weight, add it to the chunk
        if total_weight[min_chunk_index] + input_dict[key] <= target_weight:
            chunks[min_chunk_index].append(key)
            total_weight[min_chunk_index] += input_dict[key]
        else:
            # Find the chunk with the weight closest to the target_weight
            closest_chunk_index = min(
                range(num_chunks),
                key=lambda i: abs(total_weight[i] + input_dict[key] - target_weight),
            )
            chunks[closest_chunk_index].append(key)
            total_weight[closest_chunk_index] += input_dict[key]

    # Close the progress bar
    progress_bar.close()

    # Initial balance
    initial_balance = max(len(chunk) for chunk in chunks) - min(
        len(chunk) for chunk in chunks
    )

    # Refinement step
    for _ in range(refinement_steps):
        chunks = refine_chunks(chunks, input_dict, target_weight)

        # Check for improvement in balance
        current_balance = max(len(chunk) for chunk in chunks) - min(
            len(chunk) for chunk in chunks
        )
        if current_balance >= initial_balance:
            break  # No improvement, exit the loop

        # Update initial balance for the next iteration
        initial_balance = current_balance

    # Print the min, max, and average weight of each chunk
    if verbose:
        for i, chunk in enumerate(chunks, 1):
            chunk_weights = [input_dict[key] for key in chunk]
            min_weight = min(chunk_weights)
            max_weight = max(chunk_weights)
            avg_weight = sum(chunk_weights) / len(chunk_weights)
            print(
                f"Chunk {i}: Total = {sum(chunk_weights)}, Min Weight = {min_weight}, Max Weight = {max_weight}, Average Weight = {avg_weight}"
            )

    return chunks


def create_empty_output_files(out_files):
    for key, value in out_files.items():
        if value is not None:
            # check if file exists, if not create it
            if os.path.exists(value):
                continue
            if (
                key == "bam_filtered"
                or key == "bam_reassigned"
                or key == "bam_reassigned"
            ):
                create_empty_bam(value)
            elif (
                key == "bam_filtered_tmp"
                or key == "bam_tmp"
                or key == "bam_tmp_sorted"
                or key == "bam_reassigned_tmp"
                or key == "bam_reassigned_sorted"
            ):
                continue
            else:
                Path(value).touch()


# function that creates an empty bam file
def create_empty_bam(output):
    """
    Create an empty bam file
    """
    header = {"HD": {"VN": "1.0", "SO": "unsorted"}}
    # Create an empty BAM file with the specified header
    with pysam.AlignmentFile(output, "wb", header=header) as outfile:
        pass


def check_values(val, minval, maxval, parser, var):
    try:
        value = float(val)
    except ValueError:
        parser.error(
            f"argument {var}: Invalid value {val}. Value has to be a 'float' between {minval} and {maxval}!"
        )
    value = float(val)
    if value < minval or value > maxval:
        parser.error(
            "argument %s: Invalid value %s. Range has to be between %s and %s!"
            % (
                var,
                value,
                minval,
                maxval,
            )
        )
    return float(value)


def check_values_auto(val, minval, maxval, parser, var):
    if val == "auto" or val is None or val == "None":
        return val
    else:
        # check if float
        try:
            val = float(val)
        except ValueError:
            parser.error(
                f"argument {var}: Invalid value {val}. Value has to be 'auto' or a 'float' between {minval} and {maxval}!"
            )
        return check_values(val, minval, maxval, parser, var)


# From: https://note.nkmk.me/en/python-check-int-float/
def is_integer(n):
    try:
        float(n)
    except ValueError:
        return False
    else:
        return float(n).is_integer()


# function to check if the input value has K, M or G suffix in it
def check_suffix(val, parser, var):
    if var == "--scale":
        units = ["K", "M"]
    else:
        units = ["K", "M", "G"]
    unit = val[-1]
    value = int(val[:-1])

    if is_integer(value) & (unit in units) & (value > 0):
        if var == "--scale":
            if unit == "K":
                val = value * 1000
            elif unit == "M":
                val = value * 1000000
            elif unit == "G":
                val = value * 1000000000
            return val
        else:
            return val
    else:
        parser.error(
            "argument %s: Invalid value %s. Has to be an integer larger than 0 with the following suffix K, M or G"
            % (var, val)
        )


def get_compression_type(filename):
    """
    Attempts to guess the compression (if any) on a file using the first few bytes.
    http://stackoverflow.com/questions/13044562
    """
    magic_dict = {
        "gz": (b"\x1f", b"\x8b", b"\x08"),
        "bz2": (b"\x42", b"\x5a", b"\x68"),
        "zip": (b"\x50", b"\x4b", b"\x03", b"\x04"),
    }
    max_len = max(len(x) for x in magic_dict)

    unknown_file = open(filename, "rb")
    file_start = unknown_file.read(max_len)
    unknown_file.close()
    compression_type = "plain"
    for file_type, magic_bytes in magic_dict.items():
        if file_start.startswith(magic_bytes):
            compression_type = file_type
    if compression_type == "bz2":
        sys.exit("Error: cannot use bzip2 format - use gzip instead")
        sys.exit("Error: cannot use zip format - use gzip instead")
    return compression_type


def get_open_func(filename):
    if get_compression_type(filename) == "gz":
        return gzip.open
    else:  # plain text
        return open


# From: https://stackoverflow.com/a/11541450
def is_valid_file(parser, arg, var):
    if not os.path.exists(arg):
        parser.error("argument %s: The file %s does not exist!" % (var, arg))
    else:
        return arg


# From https://stackoverflow.com/a/59617044/15704171
def convert_list_to_str(lst):
    n = len(lst)
    if not n:
        return ""
    if n == 1:
        return lst[0]
    return ", ".join(lst[:-1]) + f" or {lst[-1]}"


lca_ranks = [
    "superkingdom",
    "domain",
    "lineage",
    "kingdom",
    "subkingdom",
    "superphylum",
    "phylum",
    "subphylum",
    "superclass",
    "class",
    "subclass",
    "infraclass",
    "clade",
    "cohort",
    "subcohort",
    "superorder",
    "order",
    "suborder",
    "infraorder",
    "parvorder",
    "superfamily",
    "family",
    "subfamily",
    "tribe",
    "subtribe",
    "infratribe",
    "genus",
    "subgenus",
    "section",
    "series",
    "subseries",
    "subsection",
    "species",
    "species group",
    "species subgroup",
    "subspecies",
    "varietas",
    "morph",
    "subvariety",
    "forma",
    "forma specialis",
    "biotype",
    "genotype",
    "isolate",
    "pathogroup",
    "serogroup",
    "serotype",
    "strain",
]


def check_lca_ranks(val, parser, var):
    value = str(val)
    if value in lca_ranks:
        return value
    else:
        parser.error(
            f"argument {var}: Invalid value {value}. Filter has to be one of {convert_list_to_str(lca_ranks)}"
        )


defaults = {
    "min_read_length": 30,
    "min_read_count": 3,
    "min_expected_breadth_ratio": 0,
    "min_norm_entropy": 0,
    "min_norm_gini": 1.0,
    "min_avg_read_ani": 90.0,
    "min_read_ani": 90.0,
    "min_breadth": 0,
    "min_coverage_evenness": 0,
    "min_coeff_var": np.Inf,
    "min_coverage_mean": 0,
    "prefix": None,
    "sort_memory": "1G",
    "reference_lengths": None,
    "scale": 1e6,
    "chunk_size": None,
    "coverage_plots": None,
    "stats": None,
    "stats_filtered": None,
    "bam_filtered": None,
    "bam_reassigned": None,
    "knee_plot": None,
    "read_length_freqs": None,
    "read_hits_count": None,
    "tmp_dir": None,
    "reassign_iters": 25,
    "reassign_scale": 0.9,
    "reassign_match_reward": 1,
    "reassign_mismatch_penalty": -2,
    "reassign_gap_open_penalty": 5,
    "reassign_gap_extension_penalty": 2,
    "reassign_lambda": 1.33,
    "reassign_k": 0.621,
    "rank_lca": "species",
    "lca_summary": None,
}

help_msg = {
    "bam": "BAM file containing aligned reads",
    "threads": "Number of threads to use",
    "prefix": "Prefix used for the output files",
    "min_read_length": "Minimum read length",
    "min_read_count": "Minimum read count",
    "trim_ends": "Exclude n bases at the ends of the reference sequences",
    "trim_min": "Remove coverage that are below this percentile. Used for the Truncated Average Depth (TAD) calculation",
    "trim_max": "Remove coverage that are above this percentile. Used for the Truncated Average Depth (TAD) calculation",
    "min_breadth": "Minimum breadth",
    "min_expected_breadth_ratio": "Minimum expected breadth ratio",
    "min_norm_entropy": "Minimum normalized entropy",
    "min_norm_gini": "Minimum normalized Gini coefficient",
    "min_read_ani": "Minimum read ANI to keep a read",
    "min_avg_read_ani": "Minimum average read ANI",
    "min_coverage_evenness": "Minimum coverage evenness",
    "min_coeff_var": "Minimum coverage evenness calculated as SD/MEAN",
    "min_coverage_mean": "Minimum coverage mean",
    "transform_cov_evenness": "Include those references that fulfill all filtering criteria but the coverage evenness is 0",
    "sort_memory": "Set maximum memory per thread for sorting; suffix K/M/G recognized",
    "scale": "Scale taxonomic abundance by this factor; suffix K/M recognized",
    "read_length_freqs": "Save a JSON file with the read length frequencies mapped to each reference",
    "read_hits_count": "Save a TSV file with the read hits frequencies mapped to each reference",
    "stats": "Save a TSV file with the statistics for each reference",
    "stats_filtered": "Save a TSV file with the statistics for each reference after filtering",
    "bam_filtered": "Save a BAM file with the references that passed the filtering criteria",
    "bam_reassigned": "Save a BAM file without multimapping reads",
    "coverage_plots": "Folder where to save genome coverage plots",
    "knee_plot": "Plot knee plot",
    "sort_by_name": "Sort by read names",
    "disable_sort": "Disable sorting of the filtered BAM file",
    "chunk_size": "Chunk size for parallel processing",
    "tmp_dir": "Temporary directory",
    "help": "Help message",
    "debug": "Print debug messages",
    "reference_lengths": "File with references lengths",
    "low_memory": "Activate the low memory mode",
    "reassign": "Run an EM algorithm to reassign reads to references",
    "reassign_method": "Method for the EM algorithm",
    "reassign_iters": "Number of iterations for the EM algorithm",
    "reassign_scale": "Scale to select the best weithing alignments",
    "reassign_match_reward": "Match reward for the alignment score ",
    "reassign_mismatch_penalty": "Mismatch penalty for the alignment score ",
    "reassign_gap_open_penalty": "Gap open penalty for alignment score computation",
    "reassign_gap_extension_penalty": "Gap extension penalty for the alignment score",
    "reassign_lambda": "Lambda parameter for the alignment score",
    "reassign_k": "K parameter for the alignment score algorithm",
    "lca": "Calculate LCA for each read and estimate abundances",
    "names": "Names dmp file from taxonomy",
    "nodes": "Nodes dmp file from taxonomy",
    "acc2taxid": "acc2taxid file from taxonomy",
    "rank_lca": "Rank to use for LCA calculation",
    "lca_summary": "Save a TSV file with the LCA summary",
    "lca_missing": "Save a TSV file with references with missing taxids",
    "lca_stats": "A TSV file from the filter subcommand",
    "custom": "Use custom taxdump files",
    "version": "Print program version",
}


def get_arguments(argv=None):
    parser = argparse.ArgumentParser(
        description="A simple tool to calculate metrics from a BAM file and filter with uneven coverage.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s " + __version__,
        help=help_msg["version"],
    )
    parser.add_argument(
        "--debug", dest="debug", action="store_true", help=help_msg["debug"]
    )

    sub_parsers = parser.add_subparsers(
        help="positional arguments",
        dest="action",
    )

    # Create parent subparser. Note `add_help=False` and creation via `argparse.`
    parent_parser = argparse.ArgumentParser(add_help=False)

    required = parent_parser.add_argument_group("required arguments")
    required.add_argument(
        "--bam",
        required=True,
        dest="bam",
        type=lambda x: is_valid_file(parser, x, "bam"),
        help=help_msg["bam"],
    )
    # required = parent_parser.add_argument_group("required arguments")
    optional = parent_parser.add_argument_group("optional arguments")
    optional.add_argument(
        "-p",
        "--prefix",
        type=str,
        default=defaults["prefix"],
        metavar="STR",
        dest="prefix",
        help=help_msg["prefix"],
    )
    optional.add_argument(
        "-r",
        "--reference-lengths",
        type=lambda x: is_valid_file(parser, x, "reference_lengths"),
        metavar="FILE",
        default=defaults["reference_lengths"],
        dest="reference_lengths",
        help=help_msg["reference_lengths"],
    )
    optional.add_argument(
        "-t",
        "--threads",
        type=lambda x: int(
            check_values(x, minval=1, maxval=1000, parser=parser, var="--threads")
        ),
        dest="threads",
        metavar="INT",
        default=1,
        help=help_msg["threads"],
    )
    # Create the parser sub-command for db creation
    parser_reassign = sub_parsers.add_parser(
        "reassign",
        help="Reassign reads to references using an EM algorithm",
        parents=[parent_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    # create the parser sub-commands
    parser_filter = sub_parsers.add_parser(
        "filter",
        help="Filter references based on coverage and other metrics",
        parents=[parent_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_lca = sub_parsers.add_parser(
        "lca",
        help="Calculate LCA for each read and estimate abundances at each rank",
        parents=[parent_parser],
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    # createdb_required_args = parser_createdb.add_argument_group("required arguments")
    # reassign_required_args = parser_reassign.add_argument_group(
    #     "Re-assign required arguments"
    # )
    reassign_optional_args = parser_reassign.add_argument_group(
        "Re-assign optional arguments"
    )
    misc_reassign_args = parser_reassign.add_argument_group("miscellaneous arguments")

    filter_required_args = parser_filter.add_argument_group("Filter required arguments")
    # filter_optional_args = parser_filter.add_argument_group("Filter optional arguments")

    # lca_required_args = parser_lca.add_argument_group("LCA required arguments")
    lca_optional_args = parser_lca.add_argument_group("LCA optional arguments")

    # add subparser for filtering options:
    # reassign_args = parser.add_argument_group("reassign arguments")
    filtering_filt_args = parser_filter.add_argument_group("filtering arguments")
    # lca_args = parser.add_argument_group("lca arguments")
    misc_filter_args = parser_filter.add_argument_group("miscellaneous arguments")
    out_filter_args = parser_filter.add_argument_group("output arguments")
    # parser.add_argument(
    #     "--bam",
    #     required=True,
    #     dest="bam",
    #     type=lambda x: is_valid_file(parser, x, "bam"),
    #     help=help_msg["bam"],
    # )
    # parser.add_argument(
    #     "-t",
    #     "--threads",
    #     type=lambda x: int(
    #         check_values(x, minval=1, maxval=1000, parser=parser, var="--threads")
    #     ),
    #     dest="threads",
    #     metavar="INT",
    #     default=1,
    #     help=help_msg["threads"],
    # )

    reassign_optional_args.add_argument(
        "-i",
        "--iters",
        type=lambda x: int(
            check_values(
                x, minval=0, maxval=100000, parser=parser, var="--reassign-n-iters"
            )
        ),
        metavar="INT",
        default=defaults["reassign_iters"],
        dest="reassign_iters",
        help=help_msg["reassign_iters"],
    )
    reassign_optional_args.add_argument(
        "-s",
        "--scale",
        type=lambda x: float(
            check_values(x, minval=0, maxval=1, parser=parser, var="--scale")
        ),
        metavar="FLOAT",
        default=defaults["reassign_scale"],
        dest="reassign_scale",
        help=help_msg["reassign_scale"],
    )
    reassign_optional_args.add_argument(
        "-A",
        "--min-read-ani",
        type=lambda x: float(
            check_values(x, minval=0, maxval=100, parser=parser, var="--min-read-ani")
        ),
        metavar="FLOAT",
        default=defaults["min_read_ani"],
        dest="min_read_ani",
        help=help_msg["min_read_ani"],
    )
    reassign_optional_args.add_argument(
        "-l",
        "--min-read-length",
        type=lambda x: int(
            check_values(
                x, minval=1, maxval=100000, parser=parser, var="--min-read-length"
            )
        ),
        default=defaults["min_read_length"],
        metavar="INT",
        dest="min_read_length",
        help=help_msg["min_read_length"],
    )
    reassign_optional_args.add_argument(
        "-n",
        "--min-read-count",
        type=lambda x: int(
            check_values(
                x, minval=1, maxval=np.Inf, parser=parser, var="--min-read-count"
            )
        ),
        default=defaults["min_read_count"],
        metavar="INT",
        dest="min_read_count",
        help=help_msg["min_read_count"],
    )
    reassign_optional_args.add_argument(
        "--match-reward",
        type=lambda x: int(
            check_values(
                x, minval=0, maxval=np.Inf, parser=parser, var="--match-reward"
            )
        ),
        default=defaults["reassign_match_reward"],
        metavar="INT",
        dest="match_reward",
        help=help_msg["reassign_match_reward"],
    )
    reassign_optional_args.add_argument(
        "--mismatch-penalty",
        type=lambda x: int(
            check_values(
                x, minval=-np.Inf, maxval=0, parser=parser, var="--mismatch-penalty"
            )
        ),
        default=defaults["reassign_mismatch_penalty"],
        metavar="INT",
        dest="mismatch_penalty",
        help=help_msg["reassign_mismatch_penalty"],
    )
    reassign_optional_args.add_argument(
        "--gap-open-penalty",
        type=lambda x: int(
            check_values(
                x, minval=0, maxval=np.Inf, parser=parser, var="--gap-open-penalty"
            )
        ),
        default=defaults["reassign_gap_open_penalty"],
        metavar="INT",
        dest="gap_open_penalty",
        help=help_msg["reassign_gap_open_penalty"],
    )
    reassign_optional_args.add_argument(
        "--gap-extension-penalty",
        type=lambda x: int(
            check_values(
                x, minval=0, maxval=np.Inf, parser=parser, var="--gap-extension-penalty"
            )
        ),
        default=defaults["reassign_gap_extension_penalty"],
        metavar="INT",
        dest="gap_extension_penalty",
        help=help_msg["reassign_gap_extension_penalty"],
    )
    reassign_optional_args.add_argument(
        "--lambda",
        type=lambda x: float(
            check_values(x, minval=0, maxval=np.Inf, parser=parser, var="--lambda")
        ),
        default=defaults["reassign_lambda"],
        metavar="FLOAT",
        dest="lambda_value",
        help=help_msg["reassign_lambda"],
    )
    reassign_optional_args.add_argument(
        "-k",
        type=lambda x: float(
            check_values(x, minval=0, maxval=np.Inf, parser=parser, var="-K")
        ),
        default=defaults["reassign_k"],
        metavar="FLOAT",
        dest="K_value",
        help=help_msg["reassign_k"],
    )
    reassign_optional_args.add_argument(
        "-o",
        "--out-bam",
        dest="bam_reassigned",
        default=defaults["bam_reassigned"],
        metavar="FILE",
        type=str,
        nargs="?",
        const="",
        help=help_msg["bam_reassigned"],
    )
    reassign_optional_args.add_argument(
        "-m",
        "--sort-memory",
        type=lambda x: check_suffix(x, parser=parser, var="--sort-memory"),
        default=defaults["sort_memory"],
        metavar="STR",
        dest="sort_memory",
        help=help_msg["sort_memory"],
    )
    reassign_optional_args.add_argument(
        "-N",
        "--sort-by-name",
        dest="sort_by_name",
        action="store_true",
        help=help_msg["sort_by_name"],
    )
    reassign_optional_args.add_argument(
        "--tmp-dir",
        type=str,
        default=defaults["tmp_dir"],
        metavar="DIR",
        dest="tmp_dir",
        help=help_msg["tmp_dir"],
    )
    misc_reassign_args.add_argument(
        "--disable-sort",
        dest="disable_sort",
        action="store_true",
        help=help_msg["disable_sort"],
    )
    misc_filter_args.add_argument(
        "--reference-trim-length",
        type=lambda x: int(
            check_values(
                x, minval=0, maxval=10000, parser=parser, var="---reference-trim-length"
            )
        ),
        dest="trim_ends",
        metavar="INT",
        default=0,
        help=help_msg["trim_ends"],
    )
    misc_filter_args.add_argument(
        "--trim-min",
        type=lambda x: int(
            check_values(x, minval=0, maxval=100, parser=parser, var="--trim-min")
        ),
        dest="trim_min",
        metavar="INT",
        default=10,
        help=help_msg["trim_min"],
    )
    misc_filter_args.add_argument(
        "--trim-max",
        type=lambda x: int(
            check_values(x, minval=0, maxval=100, parser=parser, var="--trim-max")
        ),
        dest="trim_max",
        metavar="INT",
        default=90,
        help=help_msg["trim_max"],
    )
    filtering_filt_args.add_argument(
        "-A",
        "--min-read-ani",
        type=lambda x: float(
            check_values(x, minval=0, maxval=100, parser=parser, var="--min-read-ani")
        ),
        metavar="FLOAT",
        default=defaults["min_read_ani"],
        dest="min_read_ani",
        help=help_msg["min_read_ani"],
    )
    filtering_filt_args.add_argument(
        "-l",
        "--min-read-length",
        type=lambda x: int(
            check_values(
                x, minval=1, maxval=100000, parser=parser, var="--min-read-length"
            )
        ),
        default=defaults["min_read_length"],
        metavar="INT",
        dest="min_read_length",
        help=help_msg["min_read_length"],
    )
    filtering_filt_args.add_argument(
        "-n",
        "--min-read-count",
        type=lambda x: int(
            check_values(
                x, minval=1, maxval=np.Inf, parser=parser, var="--min-read-count"
            )
        ),
        default=defaults["min_read_count"],
        metavar="INT",
        dest="min_read_count",
        help=help_msg["min_read_count"],
    )
    filtering_filt_args.add_argument(
        "-b",
        "--min-expected-breadth-ratio",
        type=lambda x: float(
            check_values(
                x, minval=0, maxval=1, parser=parser, var="--min-expected-breadth-ratio"
            )
        ),
        metavar="FLOAT",
        default=defaults["min_expected_breadth_ratio"],
        dest="min_expected_breadth_ratio",
        help=help_msg["min_expected_breadth_ratio"],
    )
    filtering_filt_args.add_argument(
        "-e",
        "--min-normalized-entropy",
        type=lambda x: check_values_auto(
            x, minval=0, maxval=1, parser=parser, var="--min-normalized-entropy"
        ),
        default=defaults["min_norm_entropy"],
        metavar="FLOAT",
        dest="min_norm_entropy",
        help=help_msg["min_norm_entropy"],
    )
    filtering_filt_args.add_argument(
        "-g",
        "--min-normalized-gini",
        type=lambda x: check_values_auto(
            x, minval=0, maxval=1, parser=parser, var="--min-normalized-gini"
        ),
        default=defaults["min_norm_gini"],
        metavar="FLOAT",
        dest="min_norm_gini",
        help=help_msg["min_norm_gini"],
    )
    filtering_filt_args.add_argument(
        "-B",
        "--min-breadth",
        type=lambda x: float(
            check_values(x, minval=0, maxval=1, parser=parser, var="--min-breadth")
        ),
        default=defaults["min_breadth"],
        metavar="FLOAT",
        dest="min_breadth",
        help=help_msg["min_breadth"],
    )
    filtering_filt_args.add_argument(
        "-a",
        "--min-avg-read-ani",
        type=lambda x: float(
            check_values(
                x, minval=0, maxval=100, parser=parser, var="--min-avg-read-ani"
            )
        ),
        metavar="FLOAT",
        default=defaults["min_avg_read_ani"],
        dest="min_avg_read_ani",
        help=help_msg["min_avg_read_ani"],
    )
    filtering_filt_args.add_argument(
        "-c",
        "--min-coverage-evenness",
        type=lambda x: float(
            check_values(
                x, minval=0, maxval=1, parser=parser, var="--min-coverage-evenness"
            )
        ),
        metavar="FLOAT",
        default=defaults["min_coverage_evenness"],
        dest="min_coverage_evenness",
        help=help_msg["min_coverage_evenness"],
    )
    filtering_filt_args.add_argument(
        "-V",
        "--min-coeff-var",
        type=lambda x: float(
            check_values(
                x, minval=0, maxval=np.Inf, parser=parser, var="--min-evenness"
            )
        ),
        default=defaults["min_coeff_var"],
        metavar="FLOAT",
        dest="min_coeff_var",
        help=help_msg["min_coeff_var"],
    )
    filtering_filt_args.add_argument(
        "-C",
        "--min-coverage-mean",
        type=lambda x: float(
            check_values(
                x, minval=0, maxval=1000000, parser=parser, var="--min-coverage-mean"
            )
        ),
        default=defaults["min_coverage_mean"],
        metavar="FLOAT",
        dest="min_coverage_mean",
        help=help_msg["min_coverage_mean"],
    )
    filtering_filt_args.add_argument(
        "--include-low-detection",
        dest="transform_cov_evenness",
        action="store_true",
        help=help_msg["transform_cov_evenness"],
    )
    misc_filter_args.add_argument(
        "-m",
        "--sort-memory",
        type=lambda x: check_suffix(x, parser=parser, var="--sort-memory"),
        default=defaults["sort_memory"],
        metavar="STR",
        dest="sort_memory",
        help=help_msg["sort_memory"],
    )
    misc_filter_args.add_argument(
        "-N",
        "--sort-by-name",
        dest="sort_by_name",
        action="store_true",
        help=help_msg["sort_by_name"],
    )
    misc_filter_args.add_argument(
        "--disable-sort",
        dest="disable_sort",
        action="store_true",
        help=help_msg["disable_sort"],
    )
    misc_filter_args.add_argument(
        "--scale",
        type=lambda x: check_suffix(x, parser=parser, var="--scale"),
        default=defaults["scale"],
        dest="scale",
        metavar="STR",
        help=help_msg["scale"],
    )
    # reference_lengths
    # filter_optional_args.add_argument(
    #     "-r",
    #     "--reference-lengths",
    #     type=lambda x: is_valid_file(parser, x, "reference_lengths"),
    #     metavar="FILE",
    #     default=defaults["reference_lengths"],
    #     dest="reference_lengths",
    #     help=help_msg["reference_lengths"],
    # )
    filter_required_args.add_argument(
        "--stats",
        dest="stats",
        default=defaults["stats"],
        type=str,
        metavar="FILE",
        nargs="?",
        const="",
        required=True,
        help=help_msg["stats"],
    )
    out_filter_args.add_argument(
        "--stats-filtered",
        dest="stats_filtered",
        default=defaults["stats_filtered"],
        type=str,
        metavar="FILE",
        nargs="?",
        const="",
        help=help_msg["stats_filtered"],
    )
    out_filter_args.add_argument(
        "--bam-filtered",
        dest="bam_filtered",
        default=defaults["bam_filtered"],
        metavar="FILE",
        type=str,
        nargs="?",
        const="",
        help=help_msg["bam_filtered"],
    )
    out_filter_args.add_argument(
        "--read-length-freqs",
        dest="read_length_freqs",
        default=defaults["read_length_freqs"],
        metavar="FILE",
        type=str,
        nargs="?",
        const="",
        help=help_msg["read_length_freqs"],
    )
    out_filter_args.add_argument(
        "--read-hits-count",
        dest="read_hits_count",
        default=defaults["read_hits_count"],
        metavar="FILE",
        type=str,
        nargs="?",
        const="",
        help=help_msg["read_hits_count"],
    )
    out_filter_args.add_argument(
        "--knee-plot",
        dest="knee_plot",
        default=defaults["knee_plot"],
        metavar="FILE",
        type=str,
        nargs="?",
        const="",
        help=help_msg["knee_plot"],
    )
    out_filter_args.add_argument(
        "--coverage-plots",
        dest="coverage_plots",
        metavar="FILE",
        default=defaults["coverage_plots"],
        type=str,
        nargs="?",
        const="",
        help=help_msg["coverage_plots"],
    )
    # parser.add_argument(
    #     "--chunk-size",
    #     type=lambda x: int(
    #         check_values(x, minval=1, maxval=100000, parser=parser, var="--chunk-size")
    #     ),
    #     default=defaults["chunk_size"],
    #     metavar="INT",
    #     dest="chunk_size",
    #     help=help_msg["chunk_size"],
    # )
    misc_filter_args.add_argument(
        "--tmp-dir",
        type=str,
        default=defaults["tmp_dir"],
        metavar="DIR",
        dest="tmp_dir",
        help=help_msg["tmp_dir"],
    )
    misc_filter_args.add_argument(
        "--low-memory",
        dest="low_memory",
        action="store_true",
        help=help_msg["low_memory"],
    )

    lca_optional_args.add_argument(
        "--names",
        metavar="FILE",
        type=lambda x: is_valid_file(parser, x, "names"),
        dest="names",
        help=help_msg["names"],
    )
    lca_optional_args.add_argument(
        "--nodes",
        metavar="FILE",
        type=lambda x: is_valid_file(parser, x, "nodes"),
        dest="nodes",
        help=help_msg["nodes"],
    )
    lca_optional_args.add_argument(
        "--acc2taxid",
        metavar="FILE",
        type=lambda x: is_valid_file(parser, x, "acc2taxid"),
        dest="acc2taxid",
        help=help_msg["acc2taxid"],
    )
    lca_optional_args.add_argument(
        "--lca-rank",
        metavar="STR",
        type=lambda x: str(check_lca_ranks(x, parser=parser, var="--lca-rank")),
        default=defaults["rank_lca"],
        dest="rank_lca",
        help=help_msg["rank_lca"],
    )
    lca_optional_args.add_argument(
        "--lca-summary",
        dest="lca_summary",
        metavar="FILE",
        default=defaults["lca_summary"],
        type=str,
        nargs="?",
        const="",
        help=help_msg["lca_summary"],
    )
    lca_optional_args.add_argument(
        "--scale",
        type=lambda x: check_suffix(x, parser=parser, var="--scale"),
        default=defaults["scale"],
        dest="scale",
        metavar="STR",
        help=help_msg["scale"],
    )
    lca_optional_args.add_argument(
        "-m",
        "--sort-memory",
        type=lambda x: check_suffix(x, parser=parser, var="--sort-memory"),
        default=defaults["sort_memory"],
        metavar="STR",
        dest="sort_memory",
        help=help_msg["sort_memory"],
    )
    lca_optional_args.add_argument(
        "--custom",
        dest="custom",
        action="store_true",
        help=help_msg["custom"],
    )
    lca_optional_args.add_argument(
        "--stats",
        dest="lca_stats",
        default=defaults["stats"],
        type=str,
        metavar="FILE",
        nargs="?",
        const="",
        required=False,
        help=help_msg["lca_stats"],
    )
    args = parser.parse_args(None if sys.argv[1:] else ["-h"])
    return args


@contextmanager
def suppress_stdout():
    """A context manager that redirects stdout and stderr to devnull"""
    with open(devnull, "w") as fnull:
        with redirect_stderr(fnull) as err, redirect_stdout(fnull) as out:
            yield (err, out)


def applyParallel(dfGrouped, func, threads, parms):
    p = Pool(threads)
    func = partial(func, parms=parms)
    ret_list = tqdm.tqdm(
        p.map(func, [group for name, group in dfGrouped]),
        total=len([group for name, group in dfGrouped]),
    )
    p.close()
    p.join()
    return pd.concat(ret_list)


def fast_flatten(input_list):
    return list(chain.from_iterable(input_list))


def concat_df(frames):
    COLUMN_NAMES = frames[0].columns
    df_dict = dict.fromkeys(COLUMN_NAMES, [])
    for col in COLUMN_NAMES:
        extracted = (frame[col] for frame in frames)
        # Flatten and save to df_dict
        df_dict[col] = fast_flatten(extracted)
    df = pd.DataFrame.from_dict(df_dict)[COLUMN_NAMES]
    return df


def initializer(init_data):
    global parms
    parms = init_data


def clean_up(keep, temp_dir):
    if keep:
        logging.info("Cleaning up temporary files")
        logging.shutdown()
        shutil.rmtree(temp_dir, ignore_errors=True)


# from https://stackoverflow.com/questions/53751050/python-multiprocessing-understanding-logic-behind-chunksize/54032744#54032744
def calc_chunksize(n_workers, len_iterable, factor=4):
    """Calculate chunksize argument for Pool-methods.

    Resembles source-code within `multiprocessing.pool.Pool._map_async`.
    """
    chunksize, extra = divmod(len_iterable, n_workers * factor)
    if extra:
        chunksize += 1
    return chunksize


# def create_output_files(
#     prefix,
#     bam,
#     stats,
#     stats_filtered,
#     bam_filtered,
#     read_length_freqs,
#     read_hits_count,
#     knee_plot,
#     coverage_plots,
# ):
#     if prefix is None:
#         prefix = bam.replace(".bam", "")

#     out_files = {}
#     if stats is not None:
#         if stats == "":
#             out_files["stats"] = f"{prefix}_stats.tsv.gz"
#         else:
#             out_files["stats"] = stats
#     if stats_filtered is not None:
#         if stats_filtered == "":
#             out_files["stats_filtered"] = f"{prefix}_stats-filtered.tsv.gz"
#         else:
#             out_files["stats_filtered"] = stats_filtered
#     if bam_filtered is not None:
#         if bam_filtered == "":
#             out_files["bam_filtered"] = f"{prefix}.filtered.bam"
#         else:
#             out_files["bam_filtered"] = bam_filtered
#     if read_length_freqs is not None:
#         if read_length_freqs == "":
#             out_files["read_length_freqs"] = f"{prefix}_read-length-freqs.json"
#         else:
#             out_files["read_length_freqs"] = read_length_freqs
#     if read_hits_count is not None:
#         if read_hits_count == "":
#             out_files["read_hits_count"] = f"{prefix}_read-hits-count.tsv.gz"
#         else:
#             out_files["read_hits_count"] = read_hits_count
#     if knee_plot is not None:
#         if knee_plot == "":
#             out_files["knee_plot"] = f"{prefix}_knee-plot.png"
#         else:
#             out_files["knee_plot"] = knee_plot
#     if coverage_plots is not None:
#         if coverage_plots == "":
#             out_files["coverage_plot_dir"] = f"{prefix}_coverage-plots"
#         else:
#             out_files["coverage_plot_dir"] = coverage_plots
#     out_files["bam_filtered_tmp"] = (f"{prefix}.filtered.tmp.bam",)


#     # create output files
#     out_files = {
#         "stats": stats,
#         "stats_filtered": stats_filtered,
#         "bam_filtered_tmp": f"{prefix}.filtered.tmp.bam",
#         "bam_filtered": bam_filtered,
#         "read_length_freqs": read_length_freqs,
#         "read_hits_count": read_hits_count,
#         "knee_plot": knee_plot,
#         "coverage_plot_dir": coverage_plots,
#     }
#     return out_files
def create_output_files(
    bam,
    tmp_dir,
    prefix=None,
    mode=None,
    stats="",
    stats_filtered="",
    bam_reassigned="",
    bam_filtered="",
    read_length_freqs="",
    read_hits_count="",
    knee_plot="",
    coverage_plots="",
    lca_summary="",
):
    if prefix is None:
        prefix = Path(bam).with_suffix("").name

    if tmp_dir is not None:
        tmp_dir = tmp_dir.name
    else:
        tmp_dir = check_tmp_dir_exists(tmp_dir).name

    if stats == "" or stats is None:
        stats = f"{prefix}_stats.tsv.gz"
    if stats_filtered == "" or stats_filtered is None:
        stats_filtered = f"{prefix}_stats-filtered.tsv.gz"
    if bam_filtered == "" or bam_filtered is None:
        bam_filtered = f"{prefix}.filtered.bam"
    if bam_reassigned == "" or bam_reassigned is None:
        bam_reassigned = f"{prefix}.reassigned.bam"
    if read_length_freqs == "" or read_length_freqs is None:
        read_length_freqs = f"{prefix}_read-length-freqs.json"
    if read_hits_count == "" or read_hits_count is None:
        read_hits_count = f"{prefix}_read-hits-count.tsv.gz"
    if knee_plot == "" or knee_plot is None:
        knee_plot = f"{prefix}_knee-plot.png"
    if coverage_plots == "" or coverage_plots is None:
        coverage_plots = f"{prefix}_coverage-plots"
    if lca_summary == "" or lca_summary is None:
        lca_summary = f"{prefix}_lca-summary.tsv.gz"

    # create output files
    if mode == "filter":
        out_files = {
            "stats": stats,
            "stats_filtered": stats_filtered,
            "bam_filtered_tmp": f"{tmp_dir}/{prefix}.filtered.tmp.bam",
            "bam_filtered": bam_filtered,
            "read_length_freqs": read_length_freqs,
            "read_hits_count": read_hits_count,
            "knee_plot": knee_plot,
            "coverage_plot_dir": coverage_plots,
            "bam_tmp": f"{tmp_dir}/{prefix}.tmp.bam",
            "bam_tmp_sorted": f"{tmp_dir}/{prefix}.tmp.sorted.bam",
        }
    elif mode == "reassign":
        out_files = {
            "bam_reassigned_tmp": f"{tmp_dir}/{prefix}.reassigned.tmp.bam",
            "bam_reassigned_sorted": f"{tmp_dir}/{prefix}.reassigned.sorted.bam",
            "bam_reassigned": bam_reassigned,
        }
    elif mode == "lca":
        out_files = {
            "lca_summary": lca_summary,
        }
    else:
        log.error("Mode not recognized")
        exit(1)
    return out_files

    # out_files = {
    #     "stats": stats,
    #     "stats_filtered": stats_filtered,
    #     "bam_filtered_tmp": f"{tmp_dir}/{prefix}.filtered.tmp.bam",
    #     "bam_tmp": f"{tmp_dir}/{prefix}.tmp.bam",
    #     "bam_tmp_sorted": f"{tmp_dir}/{prefix}.tmp.sorted.bam",
    #     "bam_filtered": bam_filtered,
    #     "bam_reassigned_tmp": f"{tmp_dir}/{prefix}.reassigned.tmp.bam",
    #     "bam_reassigned_sorted": f"{tmp_dir}/{prefix}.reassigned.sorted.bam",
    #     "bam_reassigned": bam_reassigned,
    #     "read_length_freqs": read_length_freqs,
    #     "read_hits_count": read_hits_count,
    #     "knee_plot": knee_plot,
    #     "coverage_plot_dir": coverage_plots,
    #     "lca_summary": lca_summary,
    # }



import pysam
import numpy as np
import os
import sys
import pandas as pd
from multiprocessing import Pool, Manager, cpu_count
import functools
from scipy import stats
import tqdm
import logging
import warnings
from bam_filter.utils import (
    is_debug,
    calc_chunksize,
    initializer,
    create_empty_output_files,
    create_empty_bam,
    sort_keys_by_approx_weight,
)
import random
from bam_filter.entropy import entropy, norm_entropy, gini_coeff, norm_gini_coeff
from collections import defaultdict
import pyranges as pr
from pathlib import Path
import concurrent.futures
import gc

# import cProfile as profile
# import pstats

# import pstats

import matplotlib.pyplot as plt

log = logging.getLogger("my_logger")


sys.setrecursionlimit(10**6)


# def get_alns(params):
#     bam, reference = params
#     samfile = pysam.AlignmentFile(bam, "rb", threads=threads)
#     alns = []
#     for aln in samfile.fetch(
#         reference=reference, multiple_iterators=False, until_eof=True
#     ):
#         alns.append(aln.to_string())
#     return alns


def write_bam(bam, references, output_files, threads=1, sort_memory="1G"):
    logging.info("::: Writing temporary filtered BAM file... (be patient)")
    if threads > 4:
        s_threads = 4
    else:
        s_threads = threads
    samfile = pysam.AlignmentFile(bam, "rb", threads=s_threads)
    header = samfile.header
    # get read groups

    if threads > 4:
        threads = 4

    # convert the dictionary to an array
    refs_dict = dict(zip(samfile.references, samfile.lengths))
    my_array = np.array(list(refs_dict.items()))

    # get the indices of the keys to keep
    keep_indices = np.isin(my_array[:, 0], references)

    # use array indexing to get the key-value pairs to keep
    filtered_array = my_array[keep_indices]

    # convert the filtered array back to a dictionary
    refs_dict = dict(filtered_array)

    (ref_names, ref_lengths) = zip(*refs_dict.items())
    ref_lengths = list(ref_lengths)
    # convert reference lengths to integers
    ref_lengths = [int(x) for x in ref_lengths]

    refs_idx = {sys.intern(str(x)): i for i, x in enumerate(ref_names)}

    if threads > 4:
        write_threads = 4
    else:
        write_threads = threads

    out_bam_file = pysam.AlignmentFile(
        output_files["bam_tmp"],
        "wb",
        referencenames=list(ref_names),
        referencelengths=ref_lengths,
        threads=write_threads,
        header=header,
    )

    references = [x for x in samfile.references if x in refs_idx.keys()]

    logging.info(f"::: ::: Filtering {len(references):,} references sequentially...")
    for reference in tqdm.tqdm(
        references,
        total=len(references),
        leave=False,
        ncols=80,
        desc="References processed",
    ):
        for aln in samfile.fetch(
            reference=reference, multiple_iterators=False, until_eof=True
        ):
            aln.reference_id = refs_idx[aln.reference_name]
            out_bam_file.write(aln)
    out_bam_file.close()
    samfile.close()
    # prof.disable()
    # # print profiling output
    # stats = pstats.Stats(prof).strip_dirs().sort_stats("tottime")
    # stats.print_stats(5)  # top 10 rows
    if threads > 4:
        s_threads = 4
    else:
        s_threads = threads
    logging.info("::: ::: Sorting BAM file...")
    pysam.sort(
        "-@",
        str(s_threads),
        "-m",
        str(sort_memory),
        "-o",
        output_files["bam_tmp_sorted"],
        output_files["bam_tmp"],
    )

    logging.info("::: ::: BAM index not found. Indexing...")
    save = pysam.set_verbosity(0)
    if threads > 4:
        s_threads = 4
    else:
        s_threads = threads
    samfile = pysam.AlignmentFile(
        output_files["bam_tmp_sorted"], "rb", threads=s_threads
    )
    chr_lengths = []
    for chrom in samfile.references:
        chr_lengths.append(samfile.get_reference_length(chrom))
    max_chr_length = np.max(chr_lengths)
    pysam.set_verbosity(save)
    samfile.close()

    if max_chr_length > 536870912:
        logging.info("::: ::: A reference is longer than 2^29, indexing with csi")
        pysam.index(
            "-c",
            "-@",
            str(s_threads),
            output_files["bam_tmp_sorted"],
        )
    else:
        pysam.index(
            "-@",
            str(s_threads),
            output_files["bam_tmp_sorted"],
        )

    os.remove(output_files["bam_tmp"])
    return output_files["bam_tmp_sorted"]


def get_tad(cov, trim_min=10, trim_max=90):
    """
    Get the TAD of a coverage
    """
    cov = cov[
        (cov >= np.percentile(cov, trim_min)) & (cov <= np.percentile(cov, trim_max))
    ]

    if np.sum(cov) == 0:
        return 0, 0
    else:
        return np.sum(cov) / len(cov), len(cov)


# Function to calculate evenness of coverage
def coverage_evenness(coverage):
    """
    Calculate the evenness of coverage
    """
    # get coverage evenness
    # covEvenness = (
    #     (breadth * 100) / (100.0 * expBreadth * expBreadth) if expBreadth > 0 else 0
    # )
    # C = mean(X)
    # D2 = X[X<=C]
    # N = len(X)
    # n = len(D2)
    # E = 1 - (n - sum(D2) / C) / N
    C = float(np.rint(np.mean(coverage)))
    D2 = coverage[coverage <= C]
    # print(len(D2))
    # D2 = [x for x in coverage if x <= C]
    # print(len(D2))
    # exit()
    if len(D2) == 0:  # pragma: no cover
        covEvenness = 1.0
    else:
        if C > 0:
            covEvenness = 1.0 - (len(D2) - np.sum(D2) / C) / len(coverage)
        else:
            covEvenness = 0.0

    return covEvenness


# function to calculate GC content
def calc_gc_content(seq):
    """Calculate GC content of a sequence

    Args:
        seq (str): DNA sequence

    Returns:
        int: Number of GC content
    """
    gc = seq.count("G") + seq.count("C")
    return gc


def create_pyranges(reference, starts, ends, strands):
    """[summary]

    Args:
        reference ([type]): [description]
        starts ([type]): [description]
        ends ([type]): [description]
        strands ([type]): [description]
    """
    chromosomes = [reference] * len(starts)
    chromosomes = pd.Series(chromosomes).astype("category")
    starts = pd.Series(starts)
    ends = pd.Series(ends)
    strands = pd.Series(strands).astype("category")

    return pr.PyRanges(
        pd.DataFrame(
            {"Chromosome": chromosomes, "Start": starts, "End": ends, "Strand": strands}
        )
    )


def get_bam_stats(
    params,
    ref_lengths=None,
    min_read_ani=90.0,
    scale=1e6,
    trim_ends=0,
    trim_min=10,
    trim_max=90,
    plot=False,
    plots_dir="coverage-plots",
    read_length_freqs=False,
    threads=1,
):
    """
    Worker function per chromosome
    loop over a bam file and create tuple with lists containing metrics:
    two definitions of the edit distances to the reference genome scaled by aligned read length
    """
    # prof = profile.Profile()
    # prof.enable()
    bam, references = params
    results = []

    if threads > 4:
        threads = 4
    samfile = pysam.AlignmentFile(bam, "rb", threads=threads)
    read_hits = defaultdict(int)
    for reference in references:
        edit_distances = []
        # edit_distances_md = []
        ani_nm = []
        # ani_md = []
        read_length = []
        read_aligned_length = []
        read_mapq = []
        read_aln_score = []
        read_names = []
        read_gc_content = []
        n_alns = 0
        if ref_lengths is None:
            reference_length = int(samfile.get_reference_length(reference))
            bam_reference_length = reference_length
        else:
            reference_length = int(ref_lengths.loc[reference, "length"])
            bam_reference_length = int(samfile.get_reference_length(reference))

        log.debug(f"Processing reference {reference}")
        log.debug(f"Reference length: {reference_length:,}")
        log.debug(f"BAM reference length: {bam_reference_length:,}")
        starts = []
        ends = []
        strands = []
        cov_np = np.zeros(samfile.get_reference_length(reference), dtype=int)

        for aln in samfile.fetch(
            contig=reference, multiple_iterators=False, until_eof=True
        ):
            ani_read = (1 - ((aln.get_tag("NM") / aln.infer_query_length()))) * 100
            if ani_read >= min_read_ani:
                n_alns += 1
                read_hits[aln.query_name] += 1
                if aln.has_tag("AS"):
                    read_aln_score.append(aln.get_tag("AS"))
                else:
                    read_aln_score.append(np.nan)

                if aln.has_tag("AS"):
                    edit_distances.append(aln.get_tag("NM"))
                    ani_nm.append(ani_read)
                else:
                    edit_distances.append(np.nan)
                    ani_nm.append(np.nan)

                read_gc_content.append(calc_gc_content(aln.query_sequence))
                read_length.append(aln.infer_read_length())
                read_aligned_length.append(aln.query_alignment_length)
                read_mapq.append(aln.mapping_quality)
                read_names.append(aln.query_name)
                # check if strand is reverse
                if aln.is_reverse:
                    strand = "-"
                else:
                    strand = "+"
                starts.append(aln.reference_start)
                ends.append(aln.reference_end)
                strands.append(strand)

                cov_np[aln.reference_start : aln.reference_end] += 1
        if n_alns > 0:
            # get bases covered by reads pileup

            # cov_pos_raw = [
            #     (pileupcolumn.pos, pileupcolumn.n)
            #     for pileupcolumn in samfile.pileup(
            #         reference,
            #         start=None,
            #         stop=None,
            #         region=None,
            #         stepper="nofilter",
            #         min_mapping_quality=0,
            #         max_depth=100000000,
            #     )
            # ]
            if trim_ends > len(cov_np) or trim_ends * 2 > len(cov_np):
                log.warning(
                    f"Trimming ends ({trim_ends}) is larger than reference length ({len(cov_np)}). Disabling trimming."
                )
                trim_ends = 0

            if trim_ends > 0:
                cov_np = cov_np[trim_ends:-trim_ends]

            mean_coverage_trunc, mean_coverage_trunc_len = get_tad(
                cov_np,
                trim_min=trim_min,
                trim_max=trim_max,
            )

            cov_pos = cov_np[cov_np > 0]
            cov_positions = np.where(cov_np > 0)[0]
            # convert datafrane to pyranges
            ranges = create_pyranges(reference, starts, ends, strands)
            if ranges.df.shape[0] == 0:
                log.debug(f"No alignments found for {reference}")
                results.append(None)
                continue
            ranges_raw = ranges.merge(strand=False)
            ranges = ranges_raw.lengths().to_list()

            max_covered_bases = np.max(ranges)
            mean_covered_bases = np.mean(ranges)
            bases_covered = int(len(cov_pos))
            # get SD from covered bases
            cov_sd = np.std(cov_pos, ddof=1)
            cov_var = np.var(cov_pos, ddof=1)
            # get average coverage
            mean_coverage = np.sum(cov_pos) / (reference_length - (2 * trim_ends))
            mean_coverage_covered = np.sum(cov_pos) / bases_covered

            breadth = bases_covered / (reference_length - (2 * trim_ends))
            exp_breadth = 1 - np.exp(-mean_coverage)
            breadth_exp_ratio = breadth / exp_breadth

            if breadth_exp_ratio > 1:
                breadth_exp_ratio = 1.0

            # fill vector with zeros to match reference length
            # cov_pos_zeroes = np.pad(
            #     cov_pos, (0, reference_length - len(cov_pos)), "constant"
            # )
            cov_evenness = coverage_evenness(cov_np)
            gc_content = (np.sum(read_gc_content) / np.sum(read_length)) * 100
            c_v = cov_sd / mean_coverage
            d_i = cov_var / mean_coverage

            read_mapq = [np.nan if x == 255 else x for x in read_mapq]

            tax_abund_aln = round((n_alns / reference_length) * scale)
            tax_abund_read = round((len(set(read_names)) / reference_length) * scale)

            # Using the trimmed mean to estimate number of reads that map to the reference
            # This is to avoid the issue of having a very high coverage region that
            # skews the mean coverage
            # C = LN / G
            # • C stands for coverage
            # • G is the haploid genome length
            # • L is the read length
            # • N is the number of reads
            #
            if mean_coverage_trunc_len > 0 and mean_coverage_trunc > 0:
                n_reads_tad = round(
                    (reference_length * mean_coverage_trunc) / np.mean(read_length)
                )
                tax_abund_tad = round((n_reads_tad / mean_coverage_trunc_len) * scale)
            else:
                n_reads_tad = 0
                tax_abund_tad = 0
            # Analyse site distribution
            n_sites = len(cov_pos)
            genome_length = bam_reference_length
            # Site density (sites per thousand bp)
            site_density = 1000 * n_sites / genome_length

            counts, bins = np.histogram(
                cov_positions, bins="auto", range=(0, genome_length)
            )

            n_bins = len(bins)

            entr = entropy(counts)  # Positional entropy
            norm_entr = norm_entropy(counts)  # Normalized positional entropy
            gini = gini_coeff(counts)  # Gini coefficient
            norm_gini = norm_gini_coeff(counts)  # Normalized Gini coefficient

            log.debug(f"Number of reads: {len(set(read_names)):,}")
            log.debug(f"Number of alignments: {n_alns:,}")
            log.debug(f"Bases covered: {bases_covered:,}")
            log.debug(f"Mean coverage: {mean_coverage:.2f}")
            log.debug(f"Mean coverage (truncated): {mean_coverage_trunc:.2f}")
            log.debug(f"Reference length (truncated): {mean_coverage_trunc_len:.2f}")
            log.debug(f"Mean coverage covered: {mean_coverage_covered:.2f}")
            log.debug(f"Max covered bases: {max_covered_bases:,}")
            log.debug(f"Mean covered bases: {mean_covered_bases:.2f}")
            log.debug(f"SD: {cov_sd:.2f}")
            log.debug(f"Breadth: {breadth:.2f}")
            log.debug(f"Exp. breadth: {exp_breadth:.2f}")
            log.debug(f"Breadth/exp. ratio: {breadth_exp_ratio:.2f}")
            log.debug(f"Number of bins: {n_bins}")
            log.debug(f"Site density: {site_density:.2f}")
            log.debug(f"Entropy (H): {entr:.2f}")
            log.debug(f"Normalized entropy (H*): {norm_entr:.2f}")
            log.debug(f"Gini coefficient (G): {gini:.2f}")
            log.debug(f"Normalized Gini coefficient (G*): {norm_gini:.2f}")
            log.debug(f"Cov. evenness: {cov_evenness:.2f}")
            log.debug(f"C_v: {c_v:.2f}")
            log.debug(f"D_i: {d_i:.2f}")
            log.debug(f"Mean mapq: {np.mean(read_mapq):.2f}")
            log.debug(f"GC content: {gc_content:.2f}")
            log.debug(f"Taxonomic abundance (alns): {tax_abund_aln:.2f}")
            log.debug(f"Taxonomic abundance (reads): {tax_abund_read:.2f}")
            log.debug(f"Taxonomic abundance (TAD): {tax_abund_tad:.2f}")
            log.debug(f"Number of reads (TAD): {n_reads_tad:,}")
            if plot:
                fig, ax = plt.subplots(nrows=1, ncols=1)  # create figure & 1 axis
                # infer number of bins using Freedman-Diaconis rule
                positions_cov_zeros = pd.DataFrame(
                    {"pos": range(1, bam_reference_length + 1)}
                )
                positions_cov = pd.DataFrame({"pos": cov_positions, "cov": cov_pos})
                positions_cov = positions_cov_zeros.merge(
                    positions_cov, on="pos", how="left"
                )
                positions_cov["cov"] = positions_cov["cov"].fillna(0)
                positions_cov["cov"] = positions_cov["cov"].astype(int)
                positions_cov["cov_binary"] = positions_cov["cov"].apply(
                    lambda x: 1 if x > 0 else 0
                )
                plt.plot(
                    positions_cov["pos"],
                    positions_cov["cov"],
                    color="c",
                    ms=0.5,
                )
                plt.suptitle(f"{reference}")
                plt.title(
                    f"cov:{mean_coverage:.4f} b/e:{breadth_exp_ratio:.2f} cov_e:{cov_evenness:.2f} entropy:{norm_entr:.2f} gini:{norm_gini:.2f}"
                )
                fig.savefig(f"{plots_dir}/{reference}_coverage.png", dpi=300)
                plt.close(fig)

            data = BamAlignment(
                reference=reference,
                n_alns=n_alns,
                reference_length=reference_length,
                bam_reference_length=bam_reference_length,
                mean_coverage=mean_coverage,
                mean_coverage_trunc=mean_coverage_trunc,
                mean_coverage_trunc_len=mean_coverage_trunc_len,
                mean_coverage_covered=mean_coverage_covered,
                bases_covered=bases_covered,
                max_covered_bases=max_covered_bases,
                mean_covered_bases=mean_covered_bases,
                cov_evenness=cov_evenness,
                breadth=breadth,
                exp_breadth=exp_breadth,
                breadth_exp_ratio=breadth_exp_ratio,
                n_bins=n_bins,
                site_density=site_density,
                entropy=entr,
                norm_entropy=norm_entr,
                gini=gini,
                norm_gini=norm_gini,
                c_v=c_v,
                d_i=d_i,
                edit_distances=edit_distances,
                # edit_distances_md=edit_distances_md,
                ani_nm=ani_nm,
                # ani_md=ani_md,
                read_length=read_length,
                read_gc_content=read_gc_content,
                read_aligned_length=read_aligned_length,
                mapping_quality=read_mapq,
                read_names=set(read_names),
                read_aln_score=read_aln_score,
                tax_abund_aln=tax_abund_aln,
                tax_abund_read=tax_abund_read,
                tax_abund_tad=tax_abund_tad,
                n_reads_tad=n_reads_tad,
            )
            results.append(data)
    samfile.close()
    # prof.disable()
    # # print profiling output
    # stats = pstats.Stats(prof).strip_dirs().sort_stats("tottime")
    # stats.print_stats(5)  # top 10 rows
    # exit()
    # results = [x[0] for x in results if x[0] is not None]
    results = list(filter(None, results))
    data_df = pd.DataFrame([x.to_summary() for x in results])

    # read_hits = (
    #     pd.DataFrame.from_dict(read_hits, orient="index", columns=["count"])
    #     .rename_axis("read_id")
    #     .reset_index()
    # )
    if read_length_freqs:
        read_lens = [x.get_read_length_freqs() for x in results]
        return (data_df, read_lens, read_hits)
    else:
        return (
            data_df,
            None,
            read_hits,
        )


class BamAlignment:
    """
    Class to store alignment information
    """

    def __init__(
        self,
        reference,
        n_alns,
        read_length,
        read_gc_content,
        read_aligned_length,
        mapping_quality,
        edit_distances,
        # edit_distances_md,
        ani_nm,
        # ani_md,
        bases_covered,
        max_covered_bases,
        mean_covered_bases,
        mean_coverage,
        mean_coverage_trunc,
        mean_coverage_trunc_len,
        mean_coverage_covered,
        reference_length,
        bam_reference_length,
        breadth,
        exp_breadth,
        breadth_exp_ratio,
        n_bins,
        site_density,
        entropy,
        norm_entropy,
        gini,
        norm_gini,
        c_v,
        d_i,
        cov_evenness,
        read_names,
        read_aln_score,
        tax_abund_aln,
        tax_abund_read,
        tax_abund_tad,
        n_reads_tad,
    ):
        self.reference = reference
        self.n_alns = n_alns
        self.read_length = read_length
        self.read_gc_content = read_gc_content
        self.read_aligned_length = read_aligned_length
        self.read_aln_score = read_aln_score
        self.mapping_quality = mapping_quality
        self.edit_distances = edit_distances
        # self.edit_distances_md = edit_distances_md
        self.ani_nm = ani_nm
        # self.ani_md = ani_md
        self.bases_covered = bases_covered
        self.max_covered_bases = max_covered_bases
        self.mean_covered_bases = mean_covered_bases
        self.mean_coverage = mean_coverage
        self.mean_coverage_trunc = mean_coverage_trunc
        self.mean_coverage_trunc_len = mean_coverage_trunc_len
        self.mean_coverage_covered = mean_coverage_covered
        self.reference_length = reference_length
        self.bam_reference_length = bam_reference_length
        self.breadth = breadth
        self.exp_breadth = exp_breadth
        self.breadth_exp_ratio = breadth_exp_ratio
        self.n_bins = n_bins
        self.site_density = site_density
        self.entropy = entropy
        self.norm_entropy = norm_entropy
        self.gini = gini
        self.norm_gini = norm_gini
        self.c_v = c_v
        self.d_i = d_i
        self.cov_evenness = cov_evenness
        self.read_names = read_names
        self.tax_abund_aln = tax_abund_aln
        self.tax_abund_read = tax_abund_read
        self.tax_abund_tad = tax_abund_tad
        self.n_reads_tad = n_reads_tad
        # function to convert class to dict

    def as_dict(self):
        return {
            "reference": self.reference,
            "n_reads": self.read_names,
            "n_alns": self.n_alns,
            "read_length": self.read_length,
            "read_gc_content": self.read_gc_content,
            "read_aligned_length": self.read_aligned_length,
            "read_aln_score": self.read_aln_score,
            "mapping_quality": self.mapping_quality,
            "edit_distances": self.edit_distances,
            # "edit_distances_md": np.mean(self.edit_distances_md),
            "ani_nm": self.ani_nm,
            # "ani_md": np.mean(self.ani_md),
            "bases_covered": self.bases_covered,
            "max_covered_bases": self.max_covered_bases,
            "mean_covered_bases": self.mean_covered_bases,
            "mean_coverage": self.mean_coverage,
            "mean_coverage_trunc": self.mean_coverage_trunc,
            "mean_coverage_trunc_len": self.mean_coverage_trunc_len,
            "mean_coverage_covered": self.mean_coverage_covered,
            "reference_length": self.reference_length,
            "breadth": self.breadth,
            "exp_breadth": self.exp_breadth,
            "breadth_exp_ratio": self.breadth_exp_ratio,
            "n_bins": self.n_bins,
            "site_density": self.site_density,
            "entropy": self.entropy,
            "norm_entropy": self.norm_entropy,
            "gini": self.gini,
            "norm_gini": self.norm_gini,
            "c_v": self.c_v,
            "d_i": self.d_i,
            "cov_evenness": self.cov_evenness,
            "tax_abund_read": self.tax_abund_read,
            "tax_abund_aln": self.tax_abund_aln,
            "tax_abund_tad": self.tax_abund_tad,
            "n_reads_tad": self.n_reads_tad,
        }

    def to_summary(self):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            read_length_mean = np.mean(self.read_length)
            read_length_median = np.median(self.read_length)
            read_length_std = np.std(self.read_length, ddof=1)
            read_length_max = np.max(self.read_length)
            read_length_min = np.min(self.read_length)
            read_length_mode = stats.mode(self.read_length, keepdims=True).mode[0]
            read_aligned_length = np.mean(self.read_aligned_length)
            read_aln_score = np.mean(self.read_aln_score)
            mapping_quality = np.mean(self.mapping_quality)
            edit_distances = np.mean(self.edit_distances)
            read_ani_mean = np.mean(self.ani_nm)
            read_ani_std = np.std(self.ani_nm, ddof=1)
            read_ani_median = np.median(self.ani_nm)
            gc_content = (np.sum(self.read_gc_content) / np.sum(self.read_length)) * 100
        return {
            "reference": self.reference,
            "n_reads": len(self.read_names),
            "n_alns": self.n_alns,
            "read_length_mean": read_length_mean,
            "read_length_std": read_length_std,
            "read_length_min": read_length_min,
            "read_length_max": read_length_max,
            "read_length_median": read_length_median,
            "read_length_mode": read_length_mode,
            "gc_content": gc_content,
            "read_aligned_length": read_aligned_length,
            "read_aln_score": read_aln_score,
            "mapping_quality": mapping_quality,
            "edit_distances": edit_distances,
            # "edit_distances_md": np.mean(self.edit_distances_md),
            "read_ani_mean": read_ani_mean,
            "read_ani_std": read_ani_std,
            "read_ani_median": read_ani_median,
            # "ani_md": np.mean(self.ani_md),
            "bases_covered": self.bases_covered,
            "max_covered_bases": self.max_covered_bases,
            "mean_covered_bases": self.mean_covered_bases,
            "coverage_mean": self.mean_coverage,
            "coverage_mean_trunc": self.mean_coverage_trunc,
            "coverage_mean_trunc_len": self.mean_coverage_trunc_len,
            "coverage_covered_mean": self.mean_coverage_covered,
            "reference_length": self.reference_length,
            "bam_reference_length": self.bam_reference_length,
            "breadth": self.breadth,
            "exp_breadth": self.exp_breadth,
            "breadth_exp_ratio": self.breadth_exp_ratio,
            "n_bins": self.n_bins,
            "site_density": self.site_density,
            "entropy": self.entropy,
            "norm_entropy": self.norm_entropy,
            "gini": self.gini,
            "norm_gini": self.norm_gini,
            "c_v": self.c_v,
            "d_i": self.d_i,
            "cov_evenness": self.cov_evenness,
            "tax_abund_read": self.tax_abund_read,
            "tax_abund_aln": self.tax_abund_aln,
            "tax_abund_tad": self.tax_abund_tad,
            "n_reads_tad": self.n_reads_tad,
        }

    def get_read_length_freqs(self):
        frags = {}
        lengths = pd.Series(self.read_length)
        lengths = lengths.value_counts().sort_index()
        freqs = list(lengths / np.sum(lengths))
        frags[self.reference] = {"length": list(lengths.index), "freq": freqs}
        return frags


def check_bam_file(
    bam,
    threads=1,
    reference_lengths=None,
    sort_memory="1G",
):
    logging.info("Loading BAM file")
    save = pysam.set_verbosity(0)

    # Use a with statement to ensure proper closing of the samfile
    try:
        if threads > 4:
            s_threads = 4
        else:
            s_threads = threads

        with pysam.AlignmentFile(bam, "rb", threads=s_threads) as samfile:
            references = samfile.references

            chr_lengths = []
            for chrom in samfile.references:
                chr_lengths.append(samfile.get_reference_length(chrom))
            max_chr_length = np.max(chr_lengths)

            pysam.set_verbosity(save)
            ref_lengths = None

            if reference_lengths is not None:
                ref_lengths = pd.read_csv(
                    reference_lengths,
                    sep="\t",
                    index_col=0,
                    names=["reference", "length"],
                )
                # check if the dataframe contains all the References in the BAM file
                if not set(references).issubset(set(ref_lengths.index)):
                    logging.error(
                        "The BAM file contains references not found in the reference lengths file"
                    )
                    sys.exit(1)
                max_chr_length = np.max(ref_lengths["length"].tolist())

            # Check if BAM files is not sorted by coordinates, sort it by coordinates
            if not samfile.header["HD"]["SO"] == "coordinate":
                log.info("BAM file is not sorted by coordinates, sorting it...")
                sorted_bam = bam.replace(".bam", ".bf-sorted.bam")
                if threads > 4:
                    s_threads = 4
                else:
                    s_threads = threads
                pysam.sort(
                    "-@", str(s_threads), "-m", str(sort_memory), "-o", sorted_bam, bam
                )
                bam = sorted_bam
                pysam.index("-c", "-@", str(threads), bam)
                if threads > 4:
                    s_threads = 4
                else:
                    s_threads = threads
                samfile = pysam.AlignmentFile(bam, "rb", threads=s_threads)

            if not samfile.has_index():
                logging.info("BAM index not found. Indexing...")
                if max_chr_length > 536870912:
                    logging.info("A reference is longer than 2^29, indexing with csi")
                    pysam.index("-c", "-@", str(threads), bam)
                else:
                    pysam.index(
                        "-@",
                        str(threads),
                        bam,
                    )
            logging.info("::: BAM file looks good.")

        return bam  # No need to reload the samfile after creating index, thanks to the with statement
    except ValueError as ve:
        if "file has no sequences defined (mode='rb')" in str(ve):
            return None
        else:
            log.error(f"Message:\n{ve}")
            exit(1)


# Inspired from https://gigabaseorgigabyte.wordpress.com/2017/04/14/getting-the-edit-distance-from-a-bam-alignment-a-journey/
def process_bam(
    bam,
    threads=1,
    reference_lengths=None,
    min_read_ani=90.0,
    min_read_count=10,
    trim_ends=0,
    trim_min=10,
    trim_max=90,
    scale=1e6,
    plot=False,
    plots_dir="coverage-plots",
    read_length_freqs=False,
    output_files=None,
    low_memory=False,
    sort_memory="1G",
):
    """
    Processing function: calls pool of worker functions
    to extract from a bam file two definitions of the edit distances to the reference genome scaled by read length
    Returned in a pandas DataFrame
    """
    logging.info("Loading BAM file")
    save = pysam.set_verbosity(0)
    if threads > 4:
        s_threads = 4
    else:
        s_threads = threads
    samfile = pysam.AlignmentFile(bam, "rb", threads=s_threads)

    references = samfile.references

    # chr_lengths = []
    # for chrom in samfile.references:
    #     chr_lengths.append(samfile.get_reference_length(chrom))
    pysam.set_verbosity(save)

    ref_lengths = None
    if reference_lengths is not None:
        ref_lengths = pd.read_csv(
            reference_lengths, sep="\t", index_col=0, names=["reference", "length"]
        )
        # check if the dataframe contains all the References in the BAM file
        if not set(references).issubset(set(ref_lengths.index)):
            logging.error(
                "The BAM file contains references not found in the reference lengths file"
            )
            sys.exit(1)

    total_refs = samfile.nreferences
    logging.info(f"Found {total_refs:,} reference sequences")
    # logging.info(f"Found {samfile.mapped:,} alignments")
    # Remove references without mapped reads
    # alns_in_ref = defaultdict(int)
    # for aln in tqdm.tqdm(
    #     samfile.fetch(until_eof=True),
    #     total=samfile.mapped,
    #     leave=False,
    #     ncols=80,
    #     desc=f"Alignments processed",
    # ):
    #     alns_in_ref[aln.reference_name] += 1
    if plot:
        # Check if image folder exists
        if not os.path.exists(plots_dir):
            os.makedirs(plots_dir)

        # references = [
        #     chrom.contig
        #     for chrom in samfile.get_index_statistics()
        #     if chrom.mapped >= min_read_count
        # ]

        logging.info(f"Removing references with less than {min_read_count} reads...")

        references_m = {
            chrom.contig: chrom.mapped
            for chrom in samfile.get_index_statistics()
            if chrom.mapped >= min_read_count
        }
    else:
        logging.info(f"Removing references with less than {min_read_count} reads...")

        references_m = {
            chrom.contig: chrom.mapped
            for chrom in samfile.get_index_statistics()
            if chrom.mapped >= min_read_count
        }
        references = list(references_m.keys())

    if len(references) == 0:
        logging.warning("No reference sequences with alignments found in the BAM file")
        create_empty_output_files(output_files)
        sys.exit(0)

    logging.info(f"Keeping {len(references):,} references")

    if low_memory:
        logging.info("Low memory mode enabled, writing filtered BAM file to disk")
        samfile.close()
        bam = write_bam(
            bam=bam,
            references=references,
            output_files=output_files,
            sort_memory=sort_memory,
            threads=threads,
        )
        if threads > 4:
            s_threads = 4
        else:
            s_threads = threads
        samfile = pysam.AlignmentFile(bam, "rb", threads=s_threads)

    log.info("::: Creating reference chunks with uniform read amounts...")

    # ify the number of chunks
    ref_chunks = sort_keys_by_approx_weight(
        input_dict=references_m, scale=1, num_cores=threads, verbose=False
    )
    log.info(f"::: Created {len(ref_chunks):,} chunks")
    ref_chunks = random.sample(ref_chunks, len(ref_chunks))
    params = zip([bam] * len(ref_chunks), ref_chunks)
    try:
        logging.info(f"Processing {len(ref_chunks):,} chunks")
        if is_debug():
            data = list(
                map(
                    functools.partial(
                        get_bam_stats,
                        ref_lengths=ref_lengths,
                        min_read_ani=min_read_ani,
                        trim_ends=0,
                        trim_min=trim_min,
                        trim_max=trim_max,
                        scale=scale,
                        plot=plot,
                        plots_dir=plots_dir,
                        read_length_freqs=read_length_freqs,
                        threads=threads,
                    ),
                    params,
                )
            )
        else:
            p = Pool(
                threads,
                # initializer=initializer,
                # initargs=([params, shared_dict],),
            )

            data = list(
                tqdm.tqdm(
                    p.imap_unordered(
                        functools.partial(
                            get_bam_stats,
                            ref_lengths=ref_lengths,
                            min_read_ani=min_read_ani,
                            trim_ends=0,
                            trim_min=trim_min,
                            trim_max=trim_max,
                            scale=scale,
                            plot=plot,
                            plots_dir=plots_dir,
                            read_length_freqs=read_length_freqs,
                            threads=threads,
                        ),
                        params,
                        chunksize=1,
                    ),
                    total=len(ref_chunks),
                    leave=False,
                    ncols=80,
                    desc="References processed",
                )
            )
            p.close()
            p.join()

    except KeyboardInterrupt:
        logging.info("User canceled the operation. Terminating jobs.")
        p.terminate()
        p.join()
        sys.exit(0)
    return data


def write_to_file(alns, out_bam_file, header=None):
    for aln in alns:
        out_bam_file.write(pysam.AlignedSegment.fromstring(aln, header))


def process_references_batch(references, bam, refs_idx, min_read_ani, threads=1):
    alns = []
    if threads > 4:
        s_threads = 4
    else:
        s_threads = threads
    with pysam.AlignmentFile(bam, "rb", threads=s_threads) as samfile:
        for reference in references:
            for aln in samfile.fetch(
                reference=reference, multiple_iterators=False, until_eof=True
            ):
                ani_read = (1 - ((aln.get_tag("NM") / aln.infer_query_length()))) * 100
                if ani_read >= min_read_ani:
                    aln.reference_id = refs_idx[aln.reference_name]
                    alns.append(aln.to_string())

    return alns


def filter_reference_BAM(
    bam,
    df,
    filter_conditions,
    threads,
    out_files,
    sort_memory,
    min_read_ani,
    transform_cov_evenness=False,
    sort_by_name=False,
    disable_sort=False,
):
    """Filter BAM based on certain conditions

    Args:
        bam (str): BAM file location
        data (pandas.DataFrame): Reference statistics
        filter_conditions (dict): A dictionary with the filter conditions to be used
        out_files (dict): Where to save the BAM files.
    """
    logging.info("Filtering stats...")
    if "min_norm_entropy" in filter_conditions and "min_norm_gini" in filter_conditions:
        logging.info(
            f"::: min_read_count >= {filter_conditions['min_read_count']} "
            f"& min_read_length >= {filter_conditions['min_read_length']} "
            f"& min_avg_read_ani >= {filter_conditions['min_avg_read_ani']} "
            f"& min_expected_breadth_ratio >= {filter_conditions['min_expected_breadth_ratio']} "
            f"& min_breadth >= {filter_conditions['min_breadth']} "
            f"& min_coverage_evenness >= {filter_conditions['min_coverage_evenness']} "
            f"& min_coeff_var =< {filter_conditions['min_coeff_var']} "
            f"& min_coverage_mean >= {filter_conditions['min_coverage_mean']} "
            f"& min_norm_entropy >= {filter_conditions['min_norm_entropy']} "
            f"& min_norm_gini <= {filter_conditions['min_norm_gini']}"
        )
        # We transform the coverage_evenenness to 1.0 where the coverage is smaller than 1
        if transform_cov_evenness is True:
            df["cov_evenness_tmp"] = df["cov_evenness"]
            df["cov_evenness_tmp"] = np.where(
                np.rint(df.coverage_mean) < 1.0, 1.0, df.cov_evenness_tmp
            )
        else:
            df["cov_evenness_tmp"] = df["cov_evenness"]
        df_filtered = df.loc[
            (df["n_reads"] >= filter_conditions["min_read_count"])
            & (df["read_length_mean"] >= filter_conditions["min_read_length"])
            & (df["read_ani_mean"] >= filter_conditions["min_avg_read_ani"])
            & (
                df["breadth_exp_ratio"]
                >= filter_conditions["min_expected_breadth_ratio"]
            )
            & (df["breadth"] >= filter_conditions["min_breadth"])
            & (df["cov_evenness_tmp"] >= filter_conditions["min_coverage_evenness"])
            & (df["c_v"] <= filter_conditions["min_coeff_var"])
            & (df["coverage_mean"] >= filter_conditions["min_coverage_mean"])
            & (df["norm_entropy"] >= filter_conditions["min_norm_entropy"])
            & (df["norm_gini"] <= filter_conditions["min_norm_gini"])
        ]
    else:
        logging.info(
            f"::: min_read_count >= {filter_conditions['min_read_count']} "
            f"& min_read_length >= {filter_conditions['min_read_length']} "
            f"& min_avg_read_ani >= {filter_conditions['min_avg_read_ani']} "
            f"& min_expected_breadth_ratio >= {filter_conditions['min_expected_breadth_ratio']} "
            f"& min_breadth >= {filter_conditions['min_breadth']} "
            f"& min_coverage_evenness >= {filter_conditions['min_coverage_evenness']} "
            f"& min_coeff_var <= {filter_conditions['min_coeff_var']} "
            f"& min_coverage_mean >= {filter_conditions['min_coverage_mean']}"
        )
        if transform_cov_evenness is True:
            df["cov_evenness_tmp"] = df["cov_evenness"]
            df["cov_evenness_tmp"] = np.where(
                np.rint(df.coverage_mean) < 1.0, 1.0, df.cov_evenness_tmp
            )
        else:
            df["cov_evenness_tmp"] = df["cov_evenness"]

        df_filtered = df.loc[
            (df["n_reads"] >= filter_conditions["min_read_count"])
            & (df["read_length_mean"] >= filter_conditions["min_read_length"])
            & (df["read_ani_mean"] >= filter_conditions["min_avg_read_ani"])
            & (
                df["breadth_exp_ratio"]
                >= filter_conditions["min_expected_breadth_ratio"]
            )
            & (df["breadth"] >= filter_conditions["min_breadth"])
            & (df["cov_evenness_tmp"] >= filter_conditions["min_coverage_evenness"])
            & (df["c_v"] <= filter_conditions["min_coeff_var"])
            & (df["coverage_mean"] >= filter_conditions["min_coverage_mean"])
        ]

    del df_filtered["cov_evenness_tmp"]
    # prof = profile.Profile()
    # prof.enable()
    if len(df_filtered.index) > 0:
        # prof = profile.Profile()
        # prof.enable()
        logging.info("Saving filtered stats...")
        df_filtered.to_csv(
            out_files["stats_filtered"],
            sep="\t",
            index=False,
        )
        if out_files["bam_filtered"] is not None:
            logging.info("Writing filtered BAM file... (be patient)")
            # refs_dict = dict(
            #     zip(df_filtered["reference"], df_filtered["bam_reference_length"])
            # )
            references = df_filtered["reference"].tolist()
            if threads > 4:
                s_threads = 4
            else:
                s_threads = threads
            samfile = pysam.AlignmentFile(bam, "rb", threads=s_threads)
            refs_dict = {x: samfile.get_reference_length(x) for x in references}
            header = samfile.header
            (ref_names, ref_lengths) = zip(*refs_dict.items())
            ref_dict_m = {
                chrom.contig: chrom.mapped
                for chrom in samfile.get_index_statistics()
                if chrom.contig in ref_names
            }
            samfile.close()
            refs_idx = {sys.intern(str(x)): i for i, x in enumerate(ref_names)}
            if threads > 4:
                write_threads = 4
            else:
                write_threads = threads

            out_bam_file = pysam.AlignmentFile(
                out_files["bam_filtered_tmp"],
                "wb",
                referencenames=list(ref_names),
                referencelengths=list(ref_lengths),
                threads=write_threads,
                header=header,
            )

            # logging.info(
            #     f"::: Filtering {len(references):,} references sequentially..."
            # )
            # for reference in tqdm.tqdm(
            #     references,
            #     total=len(references),
            #     leave=False,
            #     ncols=80,
            #     desc="References processed",
            # ):
            #     for aln in samfile.fetch(
            #         reference=reference, multiple_iterators=False, until_eof=True
            #     ):
            #         ani_read = (
            #             1 - ((aln.get_tag("NM") / aln.infer_query_length()))
            #         ) * 100
            #         if ani_read >= min_read_ani:
            #             aln.reference_id = refs_idx[aln.reference_name]
            #             out_bam_file.write(aln)
            num_cores = min(threads, cpu_count())
            # batch_size = len(references) // num_cores + 1  # Ensure non-zero batch size
            log.info("::: Creating reference chunks with uniform read amounts...")
            ref_chunks = sort_keys_by_approx_weight(
                input_dict=ref_dict_m, scale=1, num_cores=threads, verbose=False
            )
            num_cores = min(num_cores, len(ref_chunks))
            log.info(f"::: Using {num_cores} cores to write {len(ref_chunks)} chunk(s)")

            # with Manager() as manager:
            #     # Use Manager to create a read-only proxy for the dictionary
            #     #refs_idx = manager.dict(refs_idx)

            with concurrent.futures.ProcessPoolExecutor(
                max_workers=num_cores
            ) as executor:
                # Use ProcessPoolExecutor to parallelize the processing of references in batches
                futures = []
                for batch_references in tqdm.tqdm(
                    ref_chunks,
                    total=len(ref_chunks),
                    desc="Submitted batches",
                    unit="batch",
                    ncols=80,
                    leave=False,
                    disable=is_debug(),
                ):
                    future = executor.submit(
                        process_references_batch,
                        batch_references,
                        bam,
                        refs_idx,
                        min_read_ani=min_read_ani,
                    )
                    futures.append(future)  # Store the future

                # Use a while loop to continuously check for completed futures
                log.info("::: Collecting batches...")

                completion_progress_bar = tqdm.tqdm(
                    total=len(futures),
                    desc="Completed",
                    unit="batch",
                    disable=is_debug(),
                    ncols=80,
                    leave=False,
                )
                completed_count = 0

                # Use as_completed to iterate over completed futures as they become available
                for completed_future in concurrent.futures.as_completed(futures):
                    alns = completed_future.result()
                    write_to_file(alns=alns, out_bam_file=out_bam_file, header=header)
                    # Update the progress bar for each completed write
                    completion_progress_bar.update(1)
                    completed_count += 1
                    completed_future.cancel()  # Cancel the future to free memory
                    gc.collect()  # Force garbage collection
                completion_progress_bar.close()
            out_bam_file.close()
            # prof.disable()
            # # print profiling output
            # stats = pstats.Stats(prof).strip_dirs().sort_stats("tottime")
            # stats.print_stats(5)  # top 10 rows
            if not disable_sort:
                if sort_by_name:
                    logging.info("Sorting BAM file by read name...")
                    if threads > 4:
                        s_threads = 4
                    else:
                        s_threads = threads
                    pysam.sort(
                        "-n",
                        "-@",
                        str(s_threads),
                        "-m",
                        str(sort_memory),
                        "-o",
                        out_files["bam_filtered"],
                        out_files["bam_filtered_tmp"],
                    )
                else:
                    logging.info("Sorting BAM file by coordinates...")
                    pysam.sort(
                        "-@",
                        str(s_threads),
                        "-m",
                        str(sort_memory),
                        "-o",
                        out_files["bam_filtered"],
                        out_files["bam_filtered_tmp"],
                    )

                    logging.info("BAM index not found. Indexing...")
                    save = pysam.set_verbosity(0)
                    if threads > 4:
                        s_threads = 4
                    else:
                        s_threads = threads
                    samfile = pysam.AlignmentFile(
                        out_files["bam_filtered"], "rb", threads=s_threads
                    )
                    chr_lengths = []
                    for chrom in samfile.references:
                        chr_lengths.append(samfile.get_reference_length(chrom))
                    max_chr_length = np.max(chr_lengths)
                    pysam.set_verbosity(save)
                    samfile.close()

                    if max_chr_length > 536870912:
                        logging.info(
                            "A reference is longer than 2^29, indexing with csi"
                        )
                        pysam.index(
                            "-c",
                            "-@",
                            str(threads),
                            out_files["bam_filtered"],
                        )
                    else:
                        pysam.index(
                            "-@",
                            str(threads),
                            out_files["bam_filtered"],
                        )

                os.remove(out_files["bam_filtered_tmp"])
            else:
                logging.info("Skipping BAM file sorting...")
                os.rename(out_files["bam_filtered_tmp"], out_files["bam_filtered"])
        else:
            logging.info("Skipping filtering BAM file creation...")
    else:
        logging.info("No references meet the filter conditions.")
        # Path(out_files["bam_filtered"]).touch()
        create_empty_bam(out_files["bam_filtered"])
        Path(out_files["stats_filtered"]).touch()
